package farm.ants.agent;

import java.rmi.*;
import java.rmi.server.*;
import java.util.*;

import farm.simulator.*;
import farm.metagent.*;
import utilities.Log;

/**
 * This is the ANTS-specific metagent, which creates the individual Agents
 * used in the simulation.  The bulk of it's functionality is derived from
 * the general FarmMetAgent, this instance basically just provides the
 * createAllAgents function to create the ANTS-specific Agent objects.  See
 * that function for more details.
 * @see #createAllAgents
 */
public class AntsMetAgent extends FarmMetAgent {
    
    // Constants
    public static final String AGENTS_DESCRIPTION = "_AgentsDescription";
    public static final String CLA_NICE = "nice";
    protected String lastdescription = null;


    /**
     * Constructor
     * @param simulator The remote simulator object to use
     * @param name The name of the metagent
     */
    public AntsMetAgent(String []args) throws RemoteException {
        super(args);

        // Command line args
        setPlugInName(getCommandLineArgument(args, CLA_NAME, "Ants MetAgent"));
    }

    /**
     * Creates the ANTS sensor and target Agents.  This is done by looking
     * for a specially formatted string at the property
     * "getMetAgentName() + AGENTS_DESCRIPTION".  The string should be of
     * the format "Sensor 0-25 Track 3-6" - a space delimited list of token
     * pairs.  The first token in each pair (e.g. Sensor or Track) indicates
     * the type of Agent to create, while the second specifies the numbers
     * of agents to create (inclusive).  So the above string would create
     * sensor agents 0 to 25, and track agents 3 to 6.  This string is
     * generally set by the FarmDriver, which is responsible for load balancing
     * and allocating the various agents among the available metagents.
     * <p>
     * The Sensor type will create SensorAgent objects, and the Track type
     * creates SPAMAgent objects.
     * @see farm.ants.plugins.FarmDriver
     * @see SensorAgent
     * @see SPAMAgent
     */
    public void createAllAgents() {

        // First flush existing agents
        stopAllAgents();

        // Then create the new ones
        String prop = getMetAgentName() + AGENTS_DESCRIPTION;
        lastdescription = (String)getProperty(prop);
        if (lastdescription != null) {
            StringTokenizer tok = new StringTokenizer(lastdescription);
            while (tok.countTokens() > 1) {
                String type = tok.nextToken();
                String range = tok.nextToken();

                int typeid = -1;
                if (type.equals("Sensor")) {
                    typeid = 0;
                } else if (type.equals("Track")) {
                    typeid = 1;
                } else if (type.equals("Optimal")) {
		    typeid = 2;
		} else if (type.equals("Random")) {
		    typeid = 3;
		}

                StringTokenizer rtok = new StringTokenizer(range, "-");
                int start = Integer.parseInt(rtok.nextToken());
                int finish = Integer.parseInt(rtok.nextToken());

                log.log("Creating " + type + " (type " + typeid + ") " + range, Log.LOG_INFO);
                for (int i = start; i <= finish; i++) {
                    switch (typeid) {
                    case 0:
                        addAgent(new SensorAgent(this, "SA" + i)); break;
                    case 1:
                        addAgent(new SPAMAgent(this, "M" + i)); break;
                    case 2:
			addAgent(new OptimalAgent(this, "O" + i)); break;
		    case 3:
			addAgent(new RandomAgent(this, "R" + i)); break;
		    }
                }
            }
        } else {
            log.log("Could not find agent description property " + prop, Log.LOG_WARNING);
        }
    }

    /**
     * This looks for the AntsMetAgent description each time it is called,
     * and if that value has changed since it was last seen, it will 
     * call createAllAgents.
     */
    protected void pulseBegin() {

        super.pulseBegin();

        // Look for a new description
        String prop = getMetAgentName() + AGENTS_DESCRIPTION;
        String agentsdesc = (String)getProperty(prop);
        if (agentsdesc != null) {
            if (lastdescription == null || !lastdescription.equals(agentsdesc)) {
                createAllAgents();
            }
        }
    }

    /**
     * The main function
     */
    public static void main(String[] args) {

        try {
            FarmMetAgent agent = new AntsMetAgent(args);
            agent.run();
        } catch (RemoteException e) {
            System.err.println("Exception: " + e.getMessage());
            e.printStackTrace();
        }
    }
}
