package farm.ants.agent;
import java.util.*;
import utilities.Log;
import farm.metagent.*; 

/**
 * An FSM implementation of the mediating portion of the SPAM
 * resource allocation protocol.
 *
 * Created: Mon Jun  4 11:33:57 2001
 *
 * @author Roger Mailler
 */
public class SPAM extends PulseActionMachine{
    int desAgents = 4;
    String name;
    Agent me = null;
    Log log = null;
    Vector locks = new Vector();

    /**
     * Constructor
     * @param m The agent this SPAM belongs to
     */
    public SPAM(Agent m) {
	me = m;
	log = Log.getDefault();
	setWaitMess(0);
	setWaitFor(WF_NONE);
	setAction("SPAM Negotiation");
	name = me.getAgentName();
    }

    public boolean executeState(int stateNum){
	switch (stateNum){
	case 0:
	    return executeState0();
	case 1:
	    return executeState1();
	case 2:
	    return executeState2();
	case 3:
	    return executeState3();
	default:
	    log.log("Unknown Track state code:" + stateNum, 
	    	    log.LOG_ERR);
	    return true;
	}
    } 

    /** 
     * This is the first state of the SPAM machine
     *
     *  
     * 1.  "Track" - the track object
     * 2.  "Sensors" - Vector Hashtable (element: String) names of 
     * sensors that could be used
     * 3.  "Timeout" - Integer deadline for negotiation completion
       
    */
    boolean executeState0() {
	Track tmpTrack = (Track) getAttribute("Track");
	Vector sensors = (Vector) getAttribute("Sensors");
	Hashtable sensScheds = requestSensorSchedules(sensors, tmpTrack);
       	int time = me.getTime();
       	int timeout = ((Integer)getAttribute("Timeout")).intValue();
	ObjFunction objective = tmpTrack.getObjFunc();
	boolean propagation = ((Boolean)getAttribute("propagate")).booleanValue();

	log.log("Executing SPAM State 0: "+ getID(), log.LOG_INFO); 

	Hashtable cleanScheds = cleanSensorScheds(sensScheds, name);

	//ensure that every sensor has a schedule
	if(cleanScheds.size() < sensors.size()){
	    for(Iterator i = sensors.iterator();i.hasNext();){
		String sens = ((SensorData) i.next()).getName();
		if(!cleanScheds.containsKey(sens)){
		    SlotSchedule SS = new SlotSchedule();
		    cleanScheds.put(sens, SS);
		}
	    }
	}
	setAttribute("Clean Schedules", cleanScheds);
	me.checkRun();

	//make sure that we can start our objective level at something
	//that has the posibility of being achieveble
	ObjEntry objE = objective.getCurrObj();
	while(objE.numAgents() > sensors.size() || objE.numAgents() > desAgents){
	    me.checkRun();
	    if(!objective.lowerObj()){
		finishSPAM();
		
		log.log("SPAM was terminated because it ran out of objective levels", log.LOG_WARNING);
		return true;
	    }
	    objE = objective.getCurrObj();
	}

	
	//reset the my niceness
	objective.highestNiceObj();
	Vector solutions =  generateConflictSol(schedulesToArray(cleanScheds), new Vector(cleanScheds.keySet()), objE, cleanScheds.size());
	while(solutions.size() == 0){
	    if(objective.lowerNiceObj()){
		log.log("Being nice, lowering to " + objective.getNiceObj(), Log.LOG_INFO);
		solutions = generateConflictSol(schedulesToArray(cleanScheds),new Vector(cleanScheds.keySet()), objective.getNiceObj(), cleanScheds.size());
		if(solutions.size() >0){
		    log.log("Found solutions, going with it ", Log.LOG_INFO);
		}
	    }
	    else{
		break;
	    }
	    me.checkRun();
	}


	boolean complete = true; 
	if(solutions.size()==0){
	    if(!propagation)
		bindIncompleteSolution(tmpTrack, sensors, cleanScheds);
	    me.checkRun();
	    complete = false;
	}
	else{
	    log.log("Binding a complete local solution", Log.LOG_INFO);
	    log.log("Binding at Objective level "+ objE.toString(), log.LOG_INFO);
	    me.checkRun();
	    Vector best = getBestSolution(solutions, sensors);
	    Hashtable currSolution = toSensorMajor(best, new Vector(), name);
	    bindSolution(tmpTrack, currSolution, objE);
	}

	
	if(!complete){
	    //going to stage 2
	    Hashtable managers = findConMans(cleanScheds, new Hashtable());

	    if(managers.size() > 0){

		sendTMSMSG(managers, sensors);
		
		//prepare the TMS for me
		TMConflict TMC = new TMConflict(name);
		for(int i=0;i<sensors.size();i++){
		    TMC.addConflict(((SensorData) sensors.elementAt(i)).getName());
		}		
		TMC.setNumAlts(sensors.size());
		TMC.getObjFunc().setCurrObj(objective.getCurrObj());
		managers.put(name, TMC);
		setAttribute("Conflicted Managers", managers);

		transition(1, WF_TIME|WF_MESSAGES, managers.size()-1, timeout);
		return false;
	    }
	}
	finishSPAM();
	return true;
    }


    boolean executeState1() {
	Vector messages = (Vector) getAttribute("Message Queue");
       	int time = me.getTime();
	Track tmpTrack = (Track) getAttribute("Track");
	Vector sensors = (Vector) getAttribute("Sensors");
       	int timeout = ((Integer)getAttribute("Timeout")).intValue();
	ObjFunction objective = tmpTrack.getObjFunc();
	Hashtable managers = (Hashtable) getAttribute("Conflicted Managers");
	Hashtable cleanScheds = (Hashtable) getAttribute("Clean Schedules");
	TrackPAM parent = (TrackPAM) getAttribute("Parent");
	
	log.log("Executing SPAM State 1: "+ getID(), log.LOG_INFO);

	//log.log("Mess " + messages.size() + " Mans " + managers.size(), log.LOG_INFO);
	if(messages.size() < managers.size()-1){
	    //didn't get messages from everybody must have timed out
	    //just exit gracefully
	    log.log("SPAM Timed Out while waiting for TMS messages" , log.LOG_INFO);
	    finishSPAM();
	    return true;
	}


	parseMetaData(messages, managers);
	me.checkRun();

	Vector mans = orderMans(managers);
	setAttribute("Conflicted Managers", mans);

	//create an empty contraint map for the function call
	Hashtable fSensInfo = new Hashtable();
	for(int i =0;i<sensors.size();i++){
	    SlotSchedule SS = new SlotSchedule();
	    fSensInfo.put(((SensorData) sensors.elementAt(i)).getName(), SS);
	}
	me.checkRun();

	boolean found = false;
	SolutionSet SS = new SolutionSet();
	log.log("Conflict: " + mans.toString(), log.LOG_INFO);
	while(!found){
	    //first let's do a sanity check, can we every expect
	    //to get a solution given the current objective levels
	    //and the sensors
	    if(solutionPossible(mans)){
		found = genGlobalSolSet(schedulesToArray(fSensInfo),new Vector(fSensInfo.keySet()), mans, 0, SS, new Solution());
	    }
	    me.checkRun();
	    if(!found){
		if(!lowerObjective(mans)){
		    tmpTrack.getObjFunc().lowestObj();
		    bindIncompleteSolution(tmpTrack, sensors, cleanScheds);
		    finishSPAM();
		    return true;
		}
	    }
	}

	Hashtable globalSolns = SS.getSolns();;
	Vector globalSolnMap = SS.getSolnsMap();
	setAttribute("Global Solutions", globalSolns);
	setAttribute("Solution Map", globalSolnMap);
	

	log.log("Requesting locks from Controller", log.LOG_DEBUG);	
	for(Iterator i = managers.keySet().iterator();i.hasNext();){
	    String manName = (String) i.next();
	    if(!manName.equals(name)){
		PACLock PACL = new PACLock(getID(), manName, me.getTime());
		log.log(PACL.toString(), log.LOG_INFO);
		locks.add(PACL);
	    }
	}
	
	me.checkRun();		
	transition(2, WF_TIME, 0, timeout, locks);
	
	return false;
    }


    boolean executeState2() {
	Vector managers = (Vector) getAttribute("Conflicted Managers");
       	int timeout = ((Integer)getAttribute("Timeout")).intValue();
	Hashtable globalSolns = (Hashtable) getAttribute("Global Solutions");

	log.log("Executing SPAM State 2: "+ getID(), log.LOG_INFO);

	Controller.printLocks();
	log.log("Requested Locks" + locks.toString(), log.LOG_INFO);
	if(Controller.numLocks(getID()) < locks.size()){
	    log.log("SPAM Timed Out while waiting for a lock", log.LOG_INFO);
	    log.log("Requested " + locks.size() + " got " + Controller.numLocks(getID()), log.LOG_INFO);
	    finishSPAM();
	    return true;	    
	}
	else{
	    setNeg(managers);
	    
	    Hashtable partSolnCnt = new Hashtable();
	    
	    sendESMSGS(managers, globalSolns, partSolnCnt, locks, timeout);
	    transition(3,WF_MESSAGES|WF_TIME, managers.size(), timeout);
	    me.checkRun();
	    setAttribute("Partial Counts", partSolnCnt);
	    return false;
	}
    }

     boolean executeState3() {
	log.log("Executing SPAM State 3: "+ getID(), log.LOG_INFO);
	//send out solution messages

	Vector messages = (Vector) getAttribute("Message Queue");
	Vector managers = (Vector) getAttribute("Conflicted Managers");
 	Vector globalSolnMap = (Vector) getAttribute("Solution Map");
	Hashtable partCounts = (Hashtable) getAttribute("Partial Counts");
	Track tmpTrack = (Track) getAttribute("Track");

	setNeg(managers);
	me.checkRun();
	if(messages.size() >= managers.size()){
	    Hashtable replyIDs = new Hashtable();
	    Hashtable orders = parseESMSGS(messages, replyIDs);
	    log.log("Orders are :" + orders.toString(), log.LOG_INFO);
	    Vector acceptSolns=chooseGlobalSoln(orders, globalSolnMap, managers, partCounts);
	    sendACMSGS(managers, acceptSolns, replyIDs);
	    tmpTrack.clearSchedules();
	}
	else{
	    log.log("SPAM Timed Out while waiting for ES responses.", log.LOG_INFO);
	}
	me.checkRun();
	finishSPAM();
	return true;
    }

    /**
     * sends out the final accept choice (AC) messages to the other
     * track managers
     */
    void sendACMSGS(Vector managers, Vector solns, Hashtable replyIDs){
	
	for(int i=0;i<managers.size();i++){
	    String man = ((TMConflict) managers.
			      elementAt(i)).getName();

	    String msg = "AC " + ((Integer) solns.elementAt(i)).intValue(); 
	    log.log("Sending " + msg + " to " + man, log.LOG_INFO); 
	    //send out the message
	    msg = (String) replyIDs.get(man)+ ":" + msg;
	    Message m = new Message(name, man, msg);
	    me.sendMessage(m);
	}
    }


    /**
     * parses and stores the response to an evaluate solutions (ES) request
     *
     * msg format: P1234:ES 0,2,3,4:P85543
     
     */
    Hashtable parseESMSGS(Vector messages, Hashtable replyIDs){
	Hashtable orders = new Hashtable();
	
	for(int i = 0;i<messages.size();i++){
	    Message m = (Message) messages.elementAt(i);
	    String sender = m.getSource();
	    String content = m.getData();	    
	    Vector vOrder = new Vector();
	    
	    StringTokenizer tok = new StringTokenizer(content,":");
	    //get past the address
	    tok.nextToken();
	    
	    //remove the ES part of the data
	    String contentData = tok.nextToken().substring(3);
	    //now get the answers
	    StringTokenizer ST = new StringTokenizer(contentData,",");
	    while(ST.hasMoreTokens()){
		vOrder.add(new Integer(ST.nextToken().trim()));
	    }
	    orders.put(sender, vOrder);
	    
	    //get the reply address
	    replyIDs.put(sender, tok.nextToken());
	}
	return orders;
    }

    /**
     * Finds the manager who is causing the over-constrained condition
     * and lower his objective function
     * @param managers A Vector of the TMConflicts for the managers in the negotiation
     * @return True if one of the managers was lowered
     */
    boolean lowerObjective(Vector managers){

	//The method for doing this is simple, just find the
	//manager with both the highest level and that cannot meet
	//his needs with his own non con sensors and lower his level

	int currMax = -1;
	for(int i=0;i<managers.size();i++){
	    TMConflict thisTM = (TMConflict) managers.elementAt(i);
	    ObjEntry obj = thisTM.getObjFunc().getCurrObj();
	    int numOwn = thisTM.getNumAlts() - thisTM.numLocalConflicts();
	    //check if eligible
	    if(obj.numAgents() > numOwn){
		if(currMax == -1){
		    //the first 
		    currMax = i;
		}
		else{
		    TMConflict maxMan = (TMConflict) managers.elementAt(currMax);
		    if(maxMan.getObjFunc().getCurrObj().compareTo(obj) > 0){
			currMax = i;
		    }
		}
	    }
	}
	if(currMax == -1){
	    log.log("Unable to lower the objective functions any more.  No eligible agents found.", log.LOG_INFO); 
	    return false;
	}

	TMConflict maxMan = (TMConflict) managers.elementAt(currMax);
	if(maxMan.getObjFunc().lowerObj()){
	    log.log("Lowering the objective function of " + maxMan.getName() + " to " + maxMan.getObjFunc().getCurrObj().toString(), log.LOG_INFO); 
	    return true;
	}
	else{
	    log.log("Unable to lower the objective functions any more, maximum manager is at lowest level", log.LOG_INFO); 
	    return false;
	}
    }

    boolean solutionPossible(Vector managers){
	int totAvail=0;
	int totDes=0;
	//log.log("total in begining " + totAvail, log.LOG_INFO);
	for(int i=0;i<managers.size();i++){
	    TMConflict thisTM = (TMConflict) managers.elementAt(i);
	    ObjEntry obj = thisTM.getObjFunc().getCurrObj();
	    if(thisTM.getName().equals(name)){
		totAvail+=thisTM.getNumAlts();
	    }
	    else{
		totAvail+=(thisTM.getNumAlts()-thisTM.numLocalConflicts());
	    }
	    totDes+=(obj.numAgents()*obj.numSlots());
	}
	//log.log("total in end " + totAvail, log.LOG_INFO);
	totAvail*=SlotSchedule.NUMSLOTS;
	return totAvail>=totDes;
    }

    /**
     * TMS reply parser
     */
    void parseMetaData(Vector messages, Hashtable managers){
    	for(int i=0;i<messages.size();i++){
	    Message m = (Message) messages.elementAt(i);
	    String content = m.getData();
	    //log.log("Got message: " + m.toString(), log.LOG_INFO);
	    //chop off the return address
	    content = content.substring(content.indexOf(":") + 1);

	    //chop off the message type id
	    content = content.substring(4);

	    StringTokenizer ST = new StringTokenizer(content, ",");
	    //int ID = Integer.parseInt(ST.nextToken());
	    String sens = m.getSource();
	    if(managers.containsKey(sens)){
		TMConflict TMC = (TMConflict) managers.get(sens);
		TMC.parseTMS(content);
		if(TMC.numLocalConflicts() > 0){
		    managers.put(sens, TMC);
		}
		else{
		    managers.remove(sens);
		}
	    }
	}
    }

    /**
     *  Request track manager state (TMS) from other track managers
     */
    void sendTMSMSG(Hashtable managers, Vector sensors){
	
	Enumeration e= managers.keys();
	while(e.hasMoreElements()){
	    String to = (String) e.nextElement();
	    String msg = "TMS ";
	    for(int i= 0;i<sensors.size()-1;i++){
		msg = msg + ((SensorData) sensors.elementAt(i)).getName() + ",";
	    }
	    msg = msg + ((SensorData) sensors.elementAt(sensors.size()-1)).getName();
	    msg = msg + ":P"+getID();

	    Message m = new Message(name, to, msg);
	    me.sendMessage(m);
	}
    }
    
    /**
     * This function creates a copy of the passed in schedules and
     * ensures that the manager passed in as killName is completly removed
     * from them.
     * @param scheds The schedules to copy
     * @param killName The name to remove from the schedules
     * @return The copy of the Hashtable
     */
    Hashtable cleanSensorScheds(Hashtable scheds, String killName){
	Hashtable HT = new Hashtable();
	Enumeration e = scheds.keys();
	while(e.hasMoreElements()){
	    String sensName = (String) e.nextElement();
	    SlotSchedule newSS = (SlotSchedule) ((SlotSchedule) scheds.get(sensName)).clone();
	    newSS.removeAll(killName);
	    HT.put(sensName, newSS);
	}
	return HT;
    }

    /**
     * generateConflictSol is responsible for, given sensor schedules,
     * determining if a non-conflicting solution exists on all or a subset
     * of the sensor and slots that are unser consideration
     * If there is then it returns a vector of all of the non-conflicting
     * solutions if none, then it return null
     * <p>
     * There are several ways to use this function that have value
     *
     * 1.  If the sensors (in schedules) have elements that are 
     * occupied, this function will take that into account.  
     * This can be used to generate the set of acceptable 
     * solutions for local resolution
     * In this use, you must specify that the availSensors = 
     * schedules.size() in order to get solutions for all of
     * the sensors
     *
     * 2.  If you want to generate the set of acceptable
     * solutiond for a set of sensors and want it to be
     * non-respective of the sensors schedules then create
     * a hashtable with the agent names as keys and having
     * empty slot schedules as elements.  Again you must
     * specify the availSensors = schedules.size()
     *
     * 3.  If you want to just get the the possible solutions
     * for 2 out of 4 sensors then schedules should only
     * contain the names and schedules of the 2 sensors you want
     * schedules for.  The availSensors should be set to 4 in this
     * case.  This will generate the combinations respective of the
     * schedules you have listed.
     *
     *
     * @param schedules key: sensore name  element: SlotSchedule
     * @param obj The current objective level
     * @return Vector (solutions) elements: Vector (slots) elements: 
     *  Vector (sensors) elements: agent names
     *     
     */
    Vector generateConflictSol(int[][] schedMap, Vector sensNames, ObjEntry obj, int availSensors) {


	//special case, he doesn't want anything
	if(obj.numAgents() * obj.numSlots() == 0){
	    Vector solution = new Vector();
	    Vector thisSoln = new Vector();
	    for(int i=0;i<SlotSchedule.NUMSLOTS;i++)
		thisSoln.add(new Vector());

	    solution.add(thisSoln);
	    return solution;
	}

	//invert the schedules
	Vector slotInfo = toSlotMajor(schedMap, sensNames);

	int goodSlots = 0;
	Vector goodStuff = new Vector();
	for(int i =0;i<slotInfo.size();i++){
	    Vector tmpVect = (Vector) slotInfo.elementAt(i);
	    int count = tmpVect.size();
	    //log.log("slot i: " + i + " count: " + count + " data: " + tmpVect.toString(), Log.LOG_DEBUG);

	    if(availSensors-(sensNames.size()-count) >= obj.numAgents()){
		goodStuff.add(new Integer(i));
		goodSlots++;
	    }
	}
	me.checkRun();

	if(goodSlots >= obj.numSlots()){
	    //great, we know there is at least one good solution
	    //log.log("Found at least one good solution.  Generating combinations", Log.LOG_SPAM);

	    //first, let's get the list of combinations of good
	    //slots, this is used in generation of the legal solutions
	    
	    Vector slotCombs = Choose(goodStuff, obj.numSlots());
	    me.checkRun();

	    //slotComb now has vectors of all possible slot combinations 
	    //which meet the criteria for a solution

	    //now we need for each of those slots, the legal combinations
	    //of agents

	    Vector sensCombs = new Vector();	
	    for(int j=0;j<slotInfo.size();j++){
		//does this slot have enough to even consider it as a 
		//a part of a viable solution?

		int numSensInSlot = ((Vector) slotInfo.elementAt(j)).size();
		int availForSlot = availSensors - (sensNames.size() - numSensInSlot);
		int lb = Math.max(0, obj.numAgents()-
				availForSlot+numSensInSlot);
		int ub = Math.min(numSensInSlot, obj.numAgents());
		//log.log("lb " + lb+ " ub " + ub, Log.LOG_SPAM);
		if(lb <= ub){
		    //the slot can generate at least one good combination
		    me.checkRun();
		    
		    Vector slotSensCombs = new Vector();
		    //for each of the slots generate the useable 
		    //sensor combination
		    
		    for(int i=lb;i<=ub;i++){
		    //generate all combinations for the n sensors such that
		    //we choose i
			Vector sensComb = Choose((Vector) slotInfo.
						 elementAt(j),i);
			slotSensCombs.addAll(sensComb);
		    }
		    sensCombs.add(slotSensCombs);
		}
		else{
		    sensCombs.add(new Vector());
		}
	    }

	    //for(int i=0;i<sensCombs.size();i++){
	    //	log.log("Combination " + i + " is " + ((Vector) sensCombs.elementAt(i)).toString(), Log.LOG_SPAM);
	    //}


	    //now we nee to construct the final answer
	    //for each slot combination, we need to pair wise match the slot 
	    //combinations in the slots
	    Vector solution = new Vector();
	    
	    for(int i =0;i<slotCombs.size();i++){
		//get one of the slot combinations
		Vector slotComb = (Vector) slotCombs.elementAt(i);
		
		//log.log("Evaluating slot combination " + i + ": " + slotComb.toString(), Log.LOG_SPAM);

		me.checkRun();

		//for this slot combination we need to construct
		//a temporary sensory combination list with only the
		//slots in the combination
		//for all the other slots, we need to add a record
		//so the table join works correctly
		
		Vector tmpSensCombs = new Vector();
		//fill tmpSensCombs with empty vectors
		for(int j =0;j<sensCombs.size();j++){
		    tmpSensCombs.add(new Vector());
		}
		
		for(int j =0;j<slotComb.size();j++){
		    int slotNum = ((Integer) slotComb.elementAt(j)).intValue();
		    tmpSensCombs.setElementAt(sensCombs.elementAt(slotNum), slotNum);
		    //log.log("Adding slot "+slotNum+" which is " + 
		    //    ((Vector) sensCombs.elementAt(slotNum)).toString()
		    //    + " to tmpSensCombs", Log.LOG_SPAM);
		}
		me.checkRun();

		Vector tmpsol = tableJoin(tmpSensCombs);

		solution.addAll(tmpsol);
	    }

	    return solution;
	}
	return new Vector();
    }

    Vector tableJoin(Vector tables){
	Vector retData = new Vector();
	maxJoin(tables, 0, new Vector(), retData);
	return retData;
    }

    void maxJoin(Vector tables, int currdepth, Vector solution, Vector Sets){
	//this function computes the pair wise join of all of the vectors
	//in the vector table.  It returns tha answer as a vector of vectors

	Vector thisTable = (Vector) tables.elementAt(currdepth);
	int maxdepth = tables.size();
	if(thisTable.size() == 0){
	    solution.add(new Vector());
	    if(currdepth < maxdepth-1){
		maxJoin(tables, currdepth +1, solution, Sets);
		solution.remove(currdepth);
	    }
	    else{
		Sets.add(new Vector(solution));
		solution.remove(currdepth);
	    }	
	}
	else{
	    for(int i=0;i<thisTable.size();i++){
		solution.add(thisTable.elementAt(i));
		if(currdepth < maxdepth-1){
		    maxJoin(tables, currdepth +1, solution, Sets);
		    solution.remove(currdepth);
		}
		else{
		    Sets.add(new Vector(solution));
		    solution.remove(currdepth);
		}
	    }
	}
    }

    Vector Choose(Vector elements, int r){
	//this function takes a vector of n elements and returns a vector
	//that contains the nCr possible combinations (vectors) of the elements

	Vector retData = new Vector();
	if(r>0){
	    nCr(r, new Vector(), 0, elements, retData);
	}
	else{
	    retData.add(new Vector());
	}
	return retData;
    }
	

    void nCr(int maxdepth, Vector solution, int first, 
	     Vector elements, Vector Sets){

	int currdepth = solution.size();
	int last = elements.size();
	for(int i = first;i<= last-maxdepth+currdepth;i++){
	    
	    solution.add(elements.elementAt(i));
	    if(currdepth < maxdepth-1){
		nCr(maxdepth, solution, i+1, elements, Sets);
		solution.remove(currdepth);
	    }
	    else{
		//Operate on solution
		Sets.add(new Vector(solution));
		solution.remove(currdepth);
	    }
	}
    }

    Hashtable toSensorMajor(Vector slotData, Vector sensors, String manData){
	/*this function returns a hashtable with each key being a sensor
	  name and each element being a SlotSchedule

	  If you specify names in a vector, the hashtable that is returned
	  contain one element per name.  If the name was not in slotdata
	  the slotschedule associated with the sensor will be empty

	  coverts        
	
       	  Slot 1 S1, S3
	  Slot 2 S2, S3
	  Slot 3 S1, S3
	  
	  to

	  S1    None  T1   None
	  S2    T1   None   T1
	  S3    None None  None

	*/

	Hashtable tmpHash = new Hashtable();
 	for(int i=0;i<sensors.size();i++){
	    SlotSchedule SS = new SlotSchedule();
	    tmpHash.put(sensors.elementAt(i),SS);
	}
	
	for(int i=0;i<slotData.size();i++){
	    Vector names = (Vector) slotData.elementAt(i);
	    for(int j = 0;j<names.size();j++){
		SlotSchedule SS;
		if(tmpHash.containsKey(names.elementAt(j))){
		    //update the current slot schedule for that
		    //sensor
		    SS = (SlotSchedule) 
			tmpHash.get(names.elementAt(j));
		    
		}
		else{
		    //create a new entry in the hashtable and
		    //set the slot schedule correctly
		    SS = new SlotSchedule();
		    tmpHash.put(names.elementAt(j), SS);
		}
		SS.add(i,manData);
	    }
	}
	return tmpHash;
    }
    void bindSolution(Track track, Hashtable solutions, ObjEntry currObjLvl){
	/*this method takes a solution in sensor major order
	  a track object and a hashtable of slot schedules
	  and updates the currrent track state with a new solution
	  as well as sends out new binding messages to enact the solution
	 */

	
	//remove agents that aren't needed any more
	Vector currSol = track.getAgents(); 
	for(int i=0;i<currSol.size();i++){
	    String thisSens = (String) currSol.elementAt(i);
	    if(!solutions.containsKey(thisSens)){
		//log.log("CurrSol " + currSol.toString());
		sendRB(thisSens, new SlotSchedule());
		track.removeAgent(thisSens);
	    }
	}
	me.checkRun();

	

	Hashtable oldAssignments = track.getAssScheds();


	Enumeration e = solutions.keys();
	while(e.hasMoreElements()){
	    String SName = (String) e.nextElement();
	    SlotSchedule SS = (SlotSchedule) solutions.get(SName);
	    SlotSchedule oldSS = null;

	    if(!oldAssignments.containsKey(SName)){
		track.addAgent(SName, SS);
	    }
	    else{	      
		track.setAgentState(SName, SS);
		oldSS = (SlotSchedule) oldAssignments.get(SName);
	    }
	    me.checkRun();

	    //if(oldSS == null || !oldSS.equals(SS)){
		sendRB(SName, SS);
		//} 
	}
	track.getObjFunc().setCurrObj(currObjLvl);
    }

    Hashtable requestSensorSchedules(Vector Sens, Track tr){
	Hashtable ret = new Hashtable();
	for(Iterator i = Sens.iterator();i.hasNext();){
	    String sensName = ((SensorData) i.next()).getName();
	    //log.log(sensName);
	    if(tr.hasSchedule(sensName)){
		ret.put(sensName, tr.getSchedule(sensName));
	    }
	}
	return ret;
    }

    void bindIncompleteSolution(Track tmpTrack, Vector sensors, Hashtable sensScheds) {
	
	ObjFunction objective = tmpTrack.getObjFunc();
	ObjEntry objE = objective.getCurrObj();

	
	log.log("Binding an incomplete local solution", Log.LOG_DEBUG);
	log.log("Binding at Objective level "+ objE.toString(), log.LOG_DEBUG);
	
	//we should prune the list of the sensors to be the best
	//ones before generating combinations.  it will save some time.
	Vector favAgentNames = getBest(sensors, objE.numAgents());
	
	Hashtable fSensInfo = new Hashtable();
	for(int i =0;i<favAgentNames.size();i++){
	    SlotSchedule SS = new SlotSchedule();
	    fSensInfo.put(((SensorData) favAgentNames.elementAt(i)).getName(), SS);
	}
	
	//no, local solution, generate an acceptable solution
	//given our local objective function that minimizes conflict
	
	Vector solutions = generateConflictSol(schedulesToArray(fSensInfo),new Vector(fSensInfo.keySet()), objE, fSensInfo.size());
	if(sensScheds !=null){
	    solutions = findMinimalConflict(sensScheds,solutions);
	}


	//bind any solution that is available
	log.log("Size: " + solutions.size(), log.LOG_DEBUG2);
	Vector best = getBestSolution(solutions, sensors);

	//log.log("Best: " + best.toString(),log.LOG_INFO);
	Hashtable currSolution = toSensorMajor(best, new Vector(), name);
	bindSolution(tmpTrack, currSolution, objE);
    }
    



    Vector findMinimalConflict(Hashtable sensMajorScheds,Vector solutions){
	/*this function will go through a list of solutions and the schedules
	  of agents and try to find the solutions which minimizes the amount
	  of conflict with existing schedules

	  Hashtable (schedules) 
	     |
	     key: sensore name  element: SlotSchedule

	  Vector (solutions)
	     |
	     elements: Vector (slots)
	                 |
			 elements: Vector (sensors)
			             |
			             elements: agent names


	  Return format is identical to solutions vector

	*/

	log.log("Generating the minimal conflicting solution set", log.LOG_DEBUG2);
    
	Vector solns = new Vector();
	int minConflicts = Integer.MAX_VALUE; 
	for(int i=0;i<solutions.size();i++){
	    int conflicts=0;
	    Vector thisSoln = (Vector) solutions.elementAt(i);
	    for(int j = 0;j<thisSoln.size();j++){
		Vector thisSlot = (Vector) thisSoln.elementAt(j);
		for(int k=0;k<thisSlot.size();k++){
		    String thisAgent = (String) thisSlot.elementAt(k);
		    SlotSchedule SS = (SlotSchedule) sensMajorScheds.get(thisAgent);
		    conflicts += SS.numCommitments(j);
		} 
	    }
	    me.checkRun();
	    log.log("Solution " + i + " has " + conflicts + " conflicts", log.LOG_DEBUG2);
	    
	    if(conflicts < minConflicts){
		log.log("It is the new minimal conflicting solution", log.LOG_DEBUG2); 
		minConflicts = conflicts;
		solns.clear();
		solns.add(thisSoln);
	    }
	    else if(conflicts == minConflicts){
		log.log("It is another minimal conflicting solutions", log.LOG_DEBUG2);
		solns.add(thisSoln);
	    }
	}    
	me.checkRun();

	return solns;
    }

    Hashtable findConMans(Hashtable schedules, Hashtable manNames){
	/*this function just goes through the sensor schedules
	  and collects the names of all conflicting track managers
	  and the names of the sensors that they are conflicted with
	  and contructs a hashtable (keyed on TM name) with TMConflict
	  elements
	  
	  this function also constructs and adds a TMConflict for the
	  track manager doing the evaluation.  For fairness
	  
	  Hashtable (schedules) 
	      |
	      key: sensore name  element: SlotSchedule

	*/
	
	//TMConflict myTMC = new TMConflict(new String(name, ID), objective);
       	//manNames.put(new String(name,ID), myTMC);

	Enumeration e = schedules.keys();
	while(e.hasMoreElements()){
	    String sensName = (String) e.nextElement();
	    SlotSchedule SS = (SlotSchedule) schedules.get(sensName);
	    //log.log("Name: " +sensName+ " Schedule " + SS.toString(),log.LOG_INFO);
	    me.checkRun();
	    for(int i = 0;i<SlotSchedule.NUMSLOTS;i++){
		Vector commits = SS.getCommitments(i);
		for(int j = 0;j<commits.size();j++){
		    if(!manNames.containsKey(commits.elementAt(j))){
			TMConflict TMC = new TMConflict((String) commits.elementAt(j));
			manNames.put(commits.elementAt(j), TMC);
		    }
		}
	    }
	}
	
	/*e = manNames.keys();
	while(e.hasMoreElements()){
	    String man = (String) e.nextElement();
	    TMConflict TMC = (TMConflict) manNames.get(man);
	    log.log("Conflicting track manager before TMS: "+ man.getManager() + "=" + TMC.toString(), log.LOG_INFO);
	}
	*/
	
	return manNames;
    }

    Vector orderMans(Hashtable managers){
	/* this function takes a hashtable, keyed on manager name
	   of TMConfilcts and orders them into a vector based on the
	   number of contraints that the managers have
	
	   done using an insertion sort because number of elements
	   is really small in general
	*/

	Vector manList = new Vector();
	Enumeration e=managers.keys();
	while(e.hasMoreElements()){
	    int position = 0;
	    String man = (String) e.nextElement();
	    TMConflict TMC = (TMConflict) managers.get(man);
	    for(int i = 0;i<manList.size();i++){
		TMConflict TMC2 = (TMConflict) manList.elementAt(i);
		log.log("Comparing " + man + " to " + TMC2.getName(), log.LOG_SPAM);

		if(TMC.numAddConflicts() > TMC2.numAddConflicts()){
		    log.log("Track Manager: " + man+ " is more conflicted", log.LOG_SPAM);

		    position = i;
		    break;
		}
		else{
		    position++;
		}
	    }
	    log.log("position: " + position, log.LOG_SPAM);
	    manList.insertElementAt(TMC, position);
	}
	return manList;
    }

    public boolean genGlobalSolSet(int[][] constraints, Vector sensNames, Vector managers, int manNum, SolutionSet SS, Solution thisSoln){
	
	/*generate the set of feasible solutions given the contraints
	  for each of these figure out if the next track manager
	  has acceptable solutions.  If they don't prune
    
	Hashtable (constraints)
	    |
	    keys: Sensor Names
            elements: SlotSchedule

	    
        Vector (managers, order by constraints)
           |
	   elements: TMConflict

	int (manNum): the manager under consideration in the recursion


	Hashtable globalSolns  //holds the set of partial solutions that 
            |                  //lead to globally feasible solutions
	    keys: Managers names
	    elements: Vector(solutions)
	                |
			elements: Vector(slots)
			            |
				    elements: Vector(sensors)
				                |
						elements:sensor names

	*/

	Hashtable schedules = new Hashtable();
	TMConflict thisTMC = (TMConflict) managers.elementAt(manNum);
	boolean foundOne = false;

	//create a Vector with only the contraints for agents that this
	//manager has in conflict
	Vector thisConflicts = thisTMC.getConflicts();
	int[][] schedMap = new int[thisConflicts.size()][SlotSchedule.NUMSLOTS];
	for(int i = 0;i<thisConflicts.size();i++){
	    for(int j = 0;j<SlotSchedule.NUMSLOTS;j++){
		schedMap[i][j] = constraints[sensNames.indexOf(thisConflicts.elementAt(i))][j];
	    }
	}
	me.checkRun();
	
	//given these constraints, generate the set of potential 
	//partial solutions for this track manager
	Vector partSol = generateConflictSol(schedMap, thisConflicts, thisTMC.getObjFunc().getCurrObj(), thisTMC.getNumAlts());


	//if we have none, the return and say that we're over constrained
	if(partSol.size() == 0){
	    //too constrained to make a solution
	    return false;
	}

	//good, we have solutions at this level to evaluate
	if(manNum < managers.size()-1){
	    for(int i = 0;i<partSol.size();i++){
		//long startTime = System.currentTimeMillis();

		//construct the constraint set for the next lower level	
		addConstraints(constraints, sensNames, 
						       (Vector) partSol.
						       elementAt(i));
		
		ResourceSchedule RS = new ResourceSchedule();
		RS.set((Vector) partSol.elementAt(i));
		thisSoln.put(thisTMC.getName(), RS);
		if(genGlobalSolSet(constraints,sensNames, managers, manNum + 1, SS, thisSoln)){
		    //there were solutions at the next lower level
		    //add this partial solution to my list of global
		    //solutions.
		    foundOne = true;
		}
		removeConstraints(constraints, sensNames, 
						       (Vector) partSol.
						       elementAt(i));
		thisSoln.remove(thisTMC.getName());
	    }
	}
	else{
	    //go ahead and add all of the solutions to the partialSolns set
	    foundOne = true;
	    //make sure we update the global solution map
	    for(int i= 0;i<partSol.size();i++){
		me.checkRun();
		ResourceSchedule RS = new ResourceSchedule();
		RS.set((Vector) partSol.elementAt(i));
		thisSoln.put(thisTMC.getName(), RS);
		//System.out.println("Adding solution: " + thisSoln.toString());
		SS.addSolution((Solution) thisSoln.clone());
		thisSoln.remove(thisTMC.getName());
	    }
	}
	return foundOne;
    }


    void addConstraints(int[][] oldConsts, Vector sensNames, Vector newConsts){

	for(int i =0;i<newConsts.size();i++){
	    Vector thisSlot = (Vector) newConsts.elementAt(i);
	    for(int j=0;j<thisSlot.size();j++){
		String thisAgent = (String) thisSlot.elementAt(j); 
		oldConsts[sensNames.indexOf(thisAgent)][i]++;
	    }   
	}
    }

    void removeConstraints(int[][] oldConsts, Vector sensNames, Vector newConsts){
	for(int i =0;i<newConsts.size();i++){
	    Vector thisSlot = (Vector) newConsts.elementAt(i);
	    for(int j=0;j<thisSlot.size();j++){
		String thisAgent = (String) thisSlot.elementAt(j); 
		oldConsts[sensNames.indexOf(thisAgent)][i]--;
	    }   
	}
    }

    /**
     * This function converts the map and the names into
     * a list for each slot of available resources
     */
    Vector toSlotMajor(int[][] schedMap, Vector names){
	//this function returns a vector of the slots where each slot is
	//a vector that contains the names of the sensors that have the slot
	//available

	/*converts        
	
	  S1    None  T1   None
	  S2    T2   None   T2
	  S3    None None  None


	  to

	  Slot 1 S1, S3
	  Slot 2 S2, S3
	  Slot 3 S1, S3
	
	*/

	Vector slotData = new Vector(SlotSchedule.NUMSLOTS);
	//count the columner counts
	for(int i = 0;i<SlotSchedule.NUMSLOTS;i++){
	    Vector sens = new Vector();
	    for(int j = 0;j<names.size();j++){
		String sensName = (String) names.elementAt(j);
		if(schedMap[j][i] == 0){
		    //free slot
		    sens.add(sensName);
		}
	    }
	    slotData.add(sens);
	}
	return slotData;
    }

    public int[][] schedulesToArray(Hashtable scheds){
	int[][] schedMap = new int[scheds.size()][SlotSchedule.NUMSLOTS];
	Vector names =  new Vector(scheds.keySet());
	for(int i = 0;i<names.size();i++){
	    String name = (String) names.elementAt(i);
	    SlotSchedule SS = (SlotSchedule) scheds.get(name);
	    for(int j = 0;j<SlotSchedule.NUMSLOTS;j++){
		schedMap[i][j] = SS.numCommitments(j);
	    }
	}
	return schedMap;
    }

    
    /**
     * ES Request message generator
     */
    void sendESMSGS(Vector managers, Hashtable globalSolns, 
		    Hashtable partCounts, Vector locks, int timeout){
	/*this function sends out the evaluate solutions message to
	  all of the track managers in the negotiation.  It returns
	  a hashtable keyed on manager name of vectors (the orderings)

	  you must pass in the ordered list of TMconflicts for the
	  managers and the hashtbale keyed on manager name of global
	  solutions
	*/
	
	Hashtable orders = new Hashtable();
	for(int i=0;i<managers.size();i++){
	    TMConflict thisTM= (TMConflict) managers.elementAt(i);
	    String man = thisTM.getName();
	    int thisManCount =0;
	    Vector solns = (Vector) globalSolns.get(man);
	    ObjEntry obj = thisTM.getObjFunc().getCurrObj();
	    String msg = "ES " + obj.toString() +
		": ";
	    boolean head = false;
	    //get the lock for this manager
	    PACLock lock = null;
	    for (Iterator j = locks.iterator();j.hasNext();){
		PACLock pacl = (PACLock) j.next();
		if(pacl.getName().equals(man)){
		    lock = pacl;
		    break;
		}
	    }
	    for(int j=0;j<solns.size();j++){
		Hashtable thisSoln = ((ResourceSchedule) solns.
		    elementAt(j)).toSlotSchedules(thisTM.getName());

		//make sure all sensors are listed
		for (Iterator k = thisTM.getConflicts().iterator();k.hasNext();){
		    String sensName = (String) k.next();
		    if(!thisSoln.containsKey(sensName)){
			thisSoln.put(sensName, new SlotSchedule());
		    }
		}

		if(!head){
		    Enumeration e = thisSoln.keys();
		    while(e.hasMoreElements()){
			msg += (String) e.nextElement();
			if(e.hasMoreElements()){
			    msg += ",";
			}
		    }
		    msg += ":";
		    head = true;
		    
		}
		Enumeration e = thisSoln.keys();
		while(e.hasMoreElements()){
		    String name = (String) e.nextElement();
		    SlotSchedule SS = (SlotSchedule) thisSoln.get(name);
		    
		    msg = msg + SS.toShortString();
		    
		}
		thisManCount++;
	    }
	    partCounts.put(man, new Integer(thisManCount)); 

	    if(lock == null) lock = new PACLock(getID(), man, me.getTime());

	    msg = msg + ":P" + getID()+":lk " + lock.getTime() + "," + lock.getLID()+":tm "+timeout;

	    //log.log("Manager " + man + " given " + thisManCount + " possible solutions ", log.LOG_INFO);

	    //log.log("Proposing these solutions " + msg, log.LOG_INFO); 
	    //send out the message
	    Message m = new Message(name, man, msg);
	    me.sendMessage(m);
	    log.log("sending message " + m, log.LOG_INFO); 
	}
    }

    Vector chooseGlobalSoln(Hashtable orders, Vector solnMap, 
			    Vector managers, Hashtable partCounts) {
	/*this function chooses the solution to apply given the
	  ordering returned by the track managers and the map of
	  globally feasible solutions
	
	  Hashtable (orders)
	     |
	     keys: manager String
	     elements: Vector (ordering)
	                 |
			 elements: solution number


	  Vector (solnMap)
	     |
	     elements: Vector (solution)
	                 |
			 elements: partial solution number

	  Vector (managers)
             |
	     elements: TMConflict

	  Hashtable (partCounts)
             |
             keys: manager (String)
	     elements: Integer (number of solutions transmitted)

	  returns:
	  
	  Vector
	    |
	    elements: partial solution number

	*/

	//intialize a vector that is equal in size to the number of managers
	Vector bestCounts = new Vector();
        for(int i=0;i<managers.size();i++){
	    bestCounts.add(new Integer(Integer.MAX_VALUE));	    
	}


	int bestSol = -1;
	for(int i=0;i<solnMap.size();i++){
	    Vector thisSoln = (Vector) solnMap.elementAt(i);
	    log.log("Evaluating mapping " + i + ": " +thisSoln.toString(), log.LOG_DEBUG2);
	    me.checkRun();
	    for(int j=0;j<bestCounts.size();j++){
		String man = ((TMConflict) managers.
				  elementAt(j)).getName();
	        log.log("Manager " + man.toString(), log.LOG_DEBUG2);
		Vector manOrder = (Vector) orders.get(man);
		int rank = manOrder.indexOf(thisSoln.elementAt(j));
	        log.log("Manager " + man.toString() + " ranked this as " + rank, log.LOG_DEBUG2);
		int currBestCount = ((Integer) bestCounts.elementAt(j)).intValue();
		if(rank < currBestCount){
		    //check to see if it is possible given the
		    //partial counts for the other managers
		    boolean okay = true;
		    for(int k=j;k<bestCounts.size();k++){
			man = ((TMConflict) managers.
					  elementAt(k)).getName();

			int pCount = ((Integer) partCounts.get(man)).intValue();
			if(((Integer) thisSoln.elementAt(k)).intValue() >= pCount){
			    okay = false;
			    break;
			}
		    }

		    if(okay){
			//it's better, so replace the
			//rest of the best counts by the
			//ranks of the other managers
			for(int k=j;k<bestCounts.size();k++){
			    man = ((TMConflict) managers.
				   elementAt(k)).getName();
			    
			    manOrder = (Vector) orders.get(man);
			    rank = manOrder.indexOf(thisSoln.elementAt(k));
			    
			    bestCounts.setElementAt(new Integer(rank), k);
			    
			}
			bestSol = i;
		    }
                    else{
                        break;
                    }
		}
		else if(rank > currBestCount){
		    //not a better solution then we have already
		    break;
		}
	    }
	}
	return (Vector) solnMap.elementAt(bestSol);
    }

    private Vector getBest(Vector list, int num){
	//return the first num elements
	return new Vector(list.subList(0, Math.min(num,list.size())));
    }
    
    private Vector getBestSolution(Vector solutions, Vector sensors){
	Random rnd = new Random();
	return (Vector) solutions.elementAt(rnd.nextInt(solutions.size()));
    }


    private void sendRB(String who, SlotSchedule SS){
	String msg = "RB " + SS.getCommitmentsInt(name);
	Message m = new Message(name, who, msg);
	log.log("Sending rebind " + m.toString(),log.LOG_INFO);
	me.sendMessage(m);
    }

    private void finishSPAM(){
	TrackPAM parent = (TrackPAM) getAttribute("Parent");
	parent.clearWaitFor();
	for(int i =0;i<locks.size();i++){
	    Controller.unlock((PACLock) locks.elementAt(i));
	}
	log.log("Finishing SPAM " + getID(), log.LOG_INFO);
	resetNeg();
    }

    public void revokeLock(PACLock lock){
	super.revokeLock(lock);
	waitNeg();
    }

    private void waitNeg(){
	me.setProperty(name+ ":state", "WAIT");    
    }

    private void resetNeg(){
	me.setProperty(name+ ":state", "N");
    }

    private void setNeg(Vector managers){
	String mString = "";
	for(int i =0;i<managers.size();i++){
	    mString += ((TMConflict) managers.elementAt(i)).getName() + ((i<managers.size()-1)?",":"");
	}
	me.setProperty(name+ ":state", mString);
    }


} // SPAM
