package farm.ants.agent;
import farm.simulator.*;
import farm.ants.plugins.*;
import farm.metagent.*;
import java.util.*;
import utilities.Log;

/**
 * Implements a simple reource centric agent, that is used to
 * test SPAM.  Is primarily composed of a PulseActionController, which
 * continuously has a TrackPAM registered to monitor the resource needs
 * of the assigned target.
 * <p>
 * The SPAMAgent also handles some of the agent specific (not addressed 
 * to a PAM) messages, such as TMS requests (sends a simple reply)
 * and ES requests (spawns a SPAMR machine)
 * <p>
 * Created: Mon Jul 15 10:26:42 2002
 *
 * @author Roger Mailler
 */
public class SPAMAgent extends Agent {
    PulseActionController PAC = new PulseActionController(this);
    String lastRes = "";
    TrackPAM trackPAM = null;
    Track track = new Track();
    float nicePercent = 0.1f;

    /**
     * Constructor
     * @param metagent A handle to the metagent this agent belongs to
     * @param name The agents name
     */    
    public SPAMAgent(FarmMetAgent metagent, String name) {
	super(metagent, name);
	log.setLevel(Log.LOG_ERR);
	nicePercent = ((Float) getProperty(FarmDriver.CLA_NICE)).floatValue();
    }
     
    /**
     * Executed on agent startup.  This method creates a TrackPAM
     * and registers it with the PAC.  It also generates the initial
     * Track object (state holder) and the objective function for the agent
     */
    public void begin() {
	super.begin();
	track.getObjFunc().setNicePercent(nicePercent);
	trackPAM = new TrackPAM(this);
	trackPAM.setAttribute("Track", track);
	PAC.register(trackPAM);
    }

    /**
     * This method simple checks the message queue for incoming
     * messages.  If the message is addressed to a PAM, then the message
     * is sent to the PAC to handle.  If it is not addressed to a PAM
     * Then this function handles it.
     */
    public void pulse(){
	super.pulse();
        checkRun();
	//log.log("Pulse", log.LOG_INFO);
	//dispatch messages
        for (Message msg = getFirstIncomingMessage(); msg != null; msg = getFirstIncomingMessage()) {
            checkRun();
	    if(msg.getData().startsWith("P")){
		PAC.messageReceived(msg); 
	    }
	    else if(msg.getData().startsWith("RS")){
		//this is a schedule update from one of the sensors
		SlotSchedule SS = new SlotSchedule();
		SS.fromPiggyBack(msg.getData().substring(3));
		track.addSchedule(msg.getSource(),SS);
		//log.log("got message " + msg.toString(), log.LOG_INFO);

	    }
	    else if(msg.getData().startsWith("TMS")){
		handleTMS(msg);
	    }
	    else if(msg.getData().startsWith("ES")){
		handleES(msg);
	    }
        }
	//pulse the PAMS
	PAC.pulse(getTime());
	//log.log("Control returned", log.LOG_INFO);
	endRun();
    }
    
    /**
     * Handles a TMS request by constructing a TMS reply from the
     * local track object.  It sends to reply to the requestng track manager
     * @param m The TMS request message
     */
    private void handleTMS(Message m){
	String content = m.getData();
	Vector conSensors = track.getConSensors();
	Vector sensors = track.getContext();
	
	log.log("PAM recieved a TMS request " + m, log.LOG_INFO);

	StringTokenizer tok = new StringTokenizer(content, ":");
	//parse the request message
	String sensorList = tok.nextToken().substring(4); //remove the TMS
	StringTokenizer ST = new StringTokenizer(sensorList, ",");
	Vector conSens = new Vector();
	int totCons = conSensors.size();
	while(ST.hasMoreTokens()){
	    String sensor = ST.nextToken().trim();
	    if(track.isConSensor(sensor)){
		totCons--;
	    }
	    if(track.inContext(sensor)){
		conSens.add(sensor);
	    }
	}

	//parse off the reply address
	String replyAddr = tok.nextToken();

	//build the return message
	String retData = "TMS " + track.getObjFunc().getCurrObj().toString()
	    + "," + sensors.size() + "," + totCons + ",";
	for(int i=0;i<conSens.size();i++){
	    if(i==conSens.size()-1){
		retData = retData + (String) conSens.elementAt(i);
	    }
	    else{
		retData = retData + (String) conSens.elementAt(i) + ",";
	    }
	}
	retData = replyAddr + ":" + retData;

	Message mes = new Message(getAgentName(), m.getSource(), retData);
	log.log("Sending reply " + mes, log.LOG_INFO);
	sendMessage(mes);
    }


    /**
     * Handles an ES request message by spawning a SPAMR machine
     * and registering it with the PAC
     * @param m The ES request message
     */
    private void handleES(Message m){
	log.log("Got an ES message from " + m.getSource(), log.LOG_INFO);
    
	SPAMR newPAM = new SPAMR(this);
	PAC.register(newPAM);
	newPAM.setAttribute("Parent", trackPAM);
	newPAM.setAttribute("Track", track);
	newPAM.addMessage(m);
    }
} // SPAMAgent
