package farm.ants.agent;
import java.util.*;

/**
 * Primary slot based scheduled used for finite time horizon
 * scheduling
 * <p>
 * Created: Mon Apr  9 19:06:24 2001
 *
 * @author Roger Mailler
 */
public class SlotSchedule implements Cloneable{
    public static final int NUMSLOTS = 3;
    ArrayList slots[];

    /**
     * Constructor
     * @param s The number of slots in the schedule
     */
    public SlotSchedule() {
	slots = new ArrayList[NUMSLOTS];
	for(int i=0;i<NUMSLOTS;i++){
	    slots[i] = new ArrayList();
	}    
    }

    /**
     * Constructor
     * @param data The string description of the schedule
     */
    public SlotSchedule(String data){
	slots = new ArrayList[NUMSLOTS];
	for(int i=0;i<NUMSLOTS;i++){
	    slots[i] = new ArrayList();
	}    
	this.fromString(data);
    }

    /**
     * Creates and returns a string description of the current schedule
     * @return a string description of the current schedule
     */
    public String toString(){
    	String sched = "";
	for(int i=0; i<NUMSLOTS;i++){
	    ArrayList slot = slots[i];
	    String slotStr = "";
	    if(slot.size() == 0){
		slotStr = "N";
	    }
	    else{
		for(int j = 0;j<slot.size()-1;j++){
		    slotStr = slotStr + (String) slot.get(j) + "/";
		}
		slotStr = slotStr + (String) slot.get(slot.size()-1); 
	    }
	    if(i<NUMSLOTS-1){
		sched = sched + slotStr + ",";
	    }
	    else{
		sched = sched + slotStr; 
	    }
	}
	return sched;
    }

    /**
     * Creates and returns a ShortString formatted string of the current schedule
     * @return a ShortString formatted string description of the current schedule
     */
    public String toShortString(){
    	String sched = "";
	int ansNum = 0;

	for(int i=0; i<NUMSLOTS;i++){
	    if(slots[i].size() != 0){
		ansNum += (int) Math.pow(2,i); 
	    }
	}
	return ansNum + "";
    }

    /**
     * Creates and returns a PiggyBack formatted string of the current schedule
     * @return a PiggyBack formatted string description of the current schedule
     */
    public String toPiggyBack(){
	String sched = "";
	Vector mans = getTracks();
	
	if(mans.size() > 0){
	    for(int i = 0;i<mans.size() -1;i++){
		String SC = (String) mans.elementAt(i);
		sched = sched + SC + "+" 
		    +  getCommitmentsInt(SC)+"+";
	    }
	    String SC = (String) mans.elementAt(mans.size()-1);
	    sched = sched + SC + "+" 
		+  getCommitmentsInt(SC);

	    return sched;
	}
	else
	    return "N";
    }

    /**
     * Clears the current schedule and replaces it with the commitments represented in a PiggyBack formatted string.
     * @param data The PiggyBack formated string used to create the schedule
     */
    public void fromPiggyBack(String data){
	clear();
	StringTokenizer tok = new StringTokenizer(data, "+");
	//is it an empty schedule?
	if(tok.countTokens() >1){
	    while(tok.hasMoreTokens()){
		String SC = tok.nextToken();
		addInt(Integer.parseInt(tok.nextToken()),SC);
	    }
	}
    }
   
    /**
     * Clears the current schedule and replaces it with the commitments from a single agent.
     * @param val The new schedule represented in integer format
     * @param name The agent changing it's binding
     */
    public void fromInt(int val, String name){
	clear();
	addInt(val, name);
    }
    
    /**
     * Modifies the schedule for a particular name to match the
     * int represented new schedule
     * @param val The new schedule represented in integer format
     * @param name The agent changing it's binding
     */
    public void addInt(int val, String name){
  	for(int i=0; i<NUMSLOTS;i++){
	    if((((int) Math.pow(2,i)) & val)>0){
		this.add(i,name);
	    } 
	    else{
		this.remove(i, name);
	    }
	}
    }

    /**
     * Parse a string description of the schedule and assigns the slot schedule to those values
     * Note, this function changes the number of slots to accomidate the number of elements in the string
     * @param data The string description
     */
    public void fromString(String data){
	clear();
	StringTokenizer ST = new StringTokenizer(data, ",");
	int s=0;
	while(ST.hasMoreTokens()){
	    String tok = ST.nextToken().trim();
	    if(!tok.equalsIgnoreCase("N")){
		StringTokenizer ST2 = new StringTokenizer(tok, "/");
		while(ST2.hasMoreTokens()){
		    add(s,ST2.nextToken());
		}
	    }
	    s++;
	    if(s>NUMSLOTS) break;
	}
    }

    /**
     * Add a commitment to a slot in the schedule
     * @param slotnum The slot number to add the committment to
     * @param name The agent getting the commitment
     */
    public void add(int slotnum, String name){
	if(!hasCommitment(slotnum, name))
	    slots[slotnum].add(name);
    }

    /**
     * Replaces the commitments in a particular slot with those described in the vector
     * @param slotnum The slot number to add the committment to
     * @param v A list of committments for that slot
     */
    public void add(int slotnum, Vector v){
	if(v==null)
	    slots[slotnum] = new ArrayList();
	else
	    slots[slotnum] = new ArrayList(v);
	
    }

    /**
     * Adds to the commitments in a particular slot with those described in the vector
     * @param slotnum The slot number to add the committment to
     * @param v A list of committments to add to that slot
     */
    public void addAll(int slotnum, Vector v){
	for(int i=0;i<v.size();i++){
	    this.add(slotnum, (String) v.elementAt(i));
	}
    }

    /**
     * Merges the commitments of one slot schedule into this one
     * @param SS The slot schedule to merge with
     */
    public void addSS(SlotSchedule SS){
	for(int i=0;i<NUMSLOTS;i++){
	    Vector v = SS.getCommitments(i);
	    addAll(i,v);
	}
    }

    /**
     * Removes a commitment from a slot
     * @param slotnum The slot number to remove the committment from
     * @param name The agent to remove
     */
    public void remove(int slotnum, String name){
	slots[slotnum].remove(name);
    }

    /**
     * Removes all of the commitments for a particular agent
     * @param name The agent to remove
     */
    public void removeAll(String name){
	for(int i=0;i<NUMSLOTS;i++){
	    slots[i].remove(name);
	}
    }

    /**
     * Determines if a slot is empty from committments or not
     * @param slotnum The slot to check
     * @return True if the slot is empty
     */
    public boolean isEmpty(int slotnum){
	return slots[slotnum].size()==0?true:false;
    }

    /**
     * Determines if a slot has a commitment from a particular agent
     * @param slotnum The slot number to check
     * @param name The name of the agent to check for
     * @return True if the agent has a committment in that slot
     */
    public boolean hasCommitment(int slotnum, String name){
	return slots[slotnum].contains(name);
    }

    /**
     * Gets the number of commitments in a slot
     * @param slotnum The slot number to check
     * @return The number of commitments
     */
    public int numCommitments(int slotnum){
	return slots[slotnum].size();
    }

    /**
     * Gets the commitments in a slot
     * @param slotnum The slot number to check
     * @return A Vector of the commitments
     */
    public Vector getCommitments(int slotnum){
	return new Vector(slots[slotnum]);    
    }

    /**
     * Determines if the schedule contains any cobound
     * slots.  The specific slots which are cobound can
     * be determined with numCommitments
     * @see #numCommitments
     */
    public boolean hasCobinding() {

	for(int i=0; i<NUMSLOTS;i++){
	    if (numCommitments(i) > 1) return true;
	}

        return false;
    }

    /**
     * Returns the schedule for a given manager in short numeric form
     * @param name The agent to get the commitments for
     * @return The interger representation of a particular agents commitments
     */
    public int getCommitmentsInt(String name){
	int ansNum = 0;
	for(int i=0; i<NUMSLOTS;i++){
	    if(hasCommitment(i, name)){
		ansNum += (int) Math.pow(2,i); 
	    }
	}
	return ansNum;
    }

    /**
     * Returns the number of slots assigned for a given manager
     * @param name The name of the agent
     * @return The number of commitments for that agent
     */
    public int getCommitmentsSlots(String name){
	int ansNum = 0;
	for(int i=0; i<NUMSLOTS;i++){
	    if(hasCommitment(i, name)){
		ansNum ++;
	    }
	}
	return ansNum;
    }    

    /**
     * Clears the schedule
     */
    public void clear(){
	for(int i=0;i<NUMSLOTS;i++){
	    slots[i].clear();
	}    
    }

    /**
     * Clones the schedule
     * @return a duplicate of the slot schedule object
     */
    public Object clone(){
    	return new SlotSchedule(this.toString());    
    }

    /**
     * Gets a list of the agents that have commitments in the schedule
     * @return A Vector of the agent names
     */
    public Vector getTracks(){
	//returns a list of SCommits, one for each unique commitment
	//in the schedule
	Vector commits = new Vector();
	for(int i=0;i<NUMSLOTS;i++){
	    Vector comms = getCommitments(i);
	    for(int j = 0;j<comms.size();j++){
		if(!commits.contains(comms.elementAt(j))){
		    commits.add(comms.elementAt(j));
		}
	    }
	}
	return commits;
    }

    /**
     * Checks to see if two schedules are the same
     * @param o1 The object to compare to
     * @return True if the schedules are the same
     */
    public boolean equals(Object o1){
	return this.toString().equals(((SlotSchedule)o1).toString());
    }

} // SlotSchedule


