package farm.ants.plugins; 
import java.rmi.*;
import java.rmi.server.*;
import farm.simulator.*;
import java.util.*;
import java.awt.geom.*;
import farm.ants.agent.AntsMetAgent;
import utilities.Log;
import farm.ants.agent.SensorData;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
/**
 * This class is the primary resource need driver in the ANT farm simulation.
 * On startup, it creates targets and sensors and distributed them 
 * across metagents.  On beginPulse, it updates the targets locations and
 * the resources needed in order to track them.
 * <p>
 * Created: Tue Jul  9 13:32:54 2002
 *
 * <p>
 * Properties:<dl>
 * <dt><tt>-width [integer]</tt>
 * <dd>The integer width of the environement (taken to be feet default: 40)
 * <dt><tt>-height [integer]</tt>
 * <dd>The integer height of the environment (taken to be feet default: 40)
 * <dt><tt>-sensors [integer]</tt>
 * <dd>The integer number of sensors in the environment (default 10)
 * <dt><tt>-targets [integer]</tt>
 * <dd>The integer number of targets to create in the environment (default 3)
 * <dt><tt>-metagents [integer]</tt>
 * <dd>The integer number of metagents to create, used for intial load 
 *     balancing (default 1)
 * <dt><tt>-placement [random|ordered]</tt>
 * <dd>Tells the driver to create the sensors in random or ordered placement (default random)
 * <dt><tt>-speed [float]</tt>
 * <dd>The maximum movement speed of a target.  The actual speed will 
 * be a number between 0.0 and the number specified.  Targets are bounced 
 * off of walls and have a small percentage chance to change their speed 
 * and direction. (default 1.0)
 * </dl>
 *
 *
 * @author Roger Mailler and Bryan Horling
 */
public class FarmDriver extends FarmPlugIn {

    // Constants
    /**
     * Property identifier for environment width
     */
    public static final String CLA_WIDTH = "width";

    /**
     * Property identifier for environment height
     */
    public static final String CLA_HEIGHT = "height";

    /**
     * Property identifier for number of sensors in the environemt
     */
    public static final String CLA_SENSORS = "sensors";

    /**
     * Property identifier for number of targets in the environmemt
     */
    public static final String CLA_TARGETS = "targets";

    /**
     * Property identifier for number of metagent in the environmemt
     */
    public static final String CLA_METAGENTS = "metagents";

    /**
     * Property identifier for the placement method of sensors 
     * in the environment
     */
    public static final String CLA_PLACEMENT = "placement";

    /**
     * Property identifier for maximum speed of targets
     */
    public static final String CLA_TARGETSPEED = "speed";
    public static final String CLA_NICE = "nice";


    int width = 40;
    int height = 40;
    int numSensors = 10;
    int numTargets = 3;
    long seed = 1234567;
    String placement = "random";
    Random rnd = null;
    ArrayList sensors = new ArrayList();
    ArrayList targets = new ArrayList();
    float sensRange = 20f;
    long lastTime = 0;
    float targSpeed = 1.0f;
    int metaAgents = 1;
    protected JFrame window;
    protected DriverPanel driverpanel;
    boolean paused = false;

    /**
     * Constructor
     * @param args An array of command line arguments for the driver
     */
    public FarmDriver(String []args) throws RemoteException {
	super(args);

        // Command line args
        setPlugInName(getCommandLineArgument(args, CLA_NAME, "Farm Driver"));
        setWidth(Integer.valueOf(getCommandLineArgument(args, CLA_WIDTH, String.valueOf(width))));
        setHeight(Integer.valueOf(getCommandLineArgument(args, CLA_HEIGHT, String.valueOf(height))));
        setNumSensors(Integer.valueOf(getCommandLineArgument(args, CLA_SENSORS, String.valueOf(numSensors))));
        setNumTargets(Integer.valueOf(getCommandLineArgument(args, CLA_TARGETS, String.valueOf(numTargets))));
        setMetaAgents(Integer.valueOf(getCommandLineArgument(args, CLA_METAGENTS, String.valueOf(metaAgents))));
        setPlacement(getCommandLineArgument(args, CLA_PLACEMENT, String.valueOf(placement)));
        setTargetSpeed(Float.valueOf(getCommandLineArgument(args, CLA_TARGETSPEED, String.valueOf(targSpeed))));
        setNicePercent(Float.valueOf(getCommandLineArgument(args, CLA_NICE, String.valueOf(0.1f))));
        rnd = getRandom();
    }

    /**
     * Sets the environment width
     * @param i The width
     */
    public void setWidth(Integer i) {
        width = i.intValue();
        setProperty(CLA_WIDTH, i);
    }

    /**
     * Sets the environment height
     * @param i The height
     */
    public void setHeight(Integer i) {
        height = i.intValue();
        setProperty(CLA_HEIGHT, i);
    }

    /**
     * Sets the number of sensors
     * @param i The number of sensors
     */
    public void setNumSensors(Integer i) {
        numSensors = i.intValue();
        setProperty(CLA_SENSORS, i);
    }

    /**
     * Sets the number of targets
     * @param i The number of targets
     */
    public void setNumTargets(Integer i) {
        numTargets = i.intValue();
        setProperty(CLA_TARGETS, i);
    }

    /**
     * Sets the nice percentage
     * @param i The percentage
     */
    public void setNicePercent(Float f) {
        setProperty(CLA_NICE, f);
    }


    /**
     * Sets the number of metagents
     * @param i The number of metagents
     */
    public void setMetaAgents(Integer i) {
        metaAgents = i.intValue(); 
        setProperty(CLA_METAGENTS, i);
    }    

    /**
     * Sets the sensor placement type
     * @param p The type of placement
     */
    public void setPlacement(String p) {
        placement = p;
        setProperty(CLA_PLACEMENT, p);
    }

    /**
     * Sets the maximum target speed
     * @param f The maximum speed
     */
    public void setTargetSpeed(Float f) {
        targSpeed = f.floatValue();
        setProperty(CLA_TARGETSPEED, f);
	if(driverpanel != null)
	    driverpanel.setTargetText(f.floatValue());
    }

    /**
     * Changes the maximum target speed
     * @param f The maximum speed
     */
    public void resetTargetSpeed(float f) {
        setTargetSpeed(new Float(f));
	if(targets != null){
	    for(Iterator i=targets.iterator();i.hasNext();){
		TargetModel targ = (TargetModel) i.next();
		targ.setMaxSpeed(f);
	    }
	}
    }

    /**
     * Sets the paused status
     * @param p True to pause
     */
    public void setPaused(boolean p) {
        paused = p;
    }
    
    /**
     * Gets the paused status
     * @return True if paused
     */
    public boolean isPaused() {
        return paused;
    }

    /**
     * If the time has changed, it moves each of the targets
     * and recalutes the resources that are available to track them
     */
    protected void pulseBegin() throws RemoteException {
	if(getTime() != lastTime){
	    //update the targets locations
	    if(!isPaused()){
		for(Iterator i=targets.iterator();i.hasNext();){
		    TargetModel targ = (TargetModel) i.next();
		    targ.move((int) (getTime()-lastTime));
		    targ.setSensors(findSensors(targ, sensors));
		}
	    }
	    lastTime = getTime();
	}
    }

    /**
     * This is called at the end of a pulse, after the agents
     * should have completed.
     */
    protected void pulseEnd() throws RemoteException {

    }

    /**
     * Starts the plugin, creates the sensors, creates the targets,
     * assigns the sensors and targets to metagents
     */
    public void begin() throws RemoteException {
	super.begin();

	sensors = createSensors(width, height, numSensors);
	targets = createTargets(width, height, numTargets);

        // Assign the sensors
        int div = (int)Math.floor(numSensors / metaAgents);
        int mod = numSensors % metaAgents;
        for (int i = 0, start = 0; i < metaAgents; i++) {
            String metkey = "MA" + i  + AntsMetAgent.AGENTS_DESCRIPTION;
            int finish = start + div-1;
            if (mod > 0) {
                finish++;
                mod--;
            }
            if (finish >= start) {
                String assign = "Sensor " + start + "-" + finish;
                String prev = (String)getProperty(metkey);
                if (prev != null)
                    assign = prev + " " + assign;
                setProperty(metkey, assign);
            }
            start = finish + 1;
         }

        // Assign the targets
        div = (int)Math.floor(numTargets / metaAgents);
        mod = numTargets % metaAgents;
        for (int i = metaAgents-1, start = 0; i >= 0; i--) {
            String metkey = "MA" + i  + AntsMetAgent.AGENTS_DESCRIPTION;
            int finish = start + div-1;
            if (mod > 0) {
                finish++;
                mod--;
            }
            if (finish >= start) {
                String assign = "Track " + start + "-" + finish;
                String prev = (String)getProperty(metkey);
                if (prev != null)
                    assign = prev + " " + assign;
                setProperty(metkey, assign);
            }
            start = finish + 1;
        }

        String assign = "";
	String metkey = "MA0" + AntsMetAgent.AGENTS_DESCRIPTION;

	//create an optimal agent for comparison
	//assign =  "Optimal 0-0";
	
	//create an random agent for comparison  
	assign = assign + " Random 0-0";

	String prev = (String)getProperty(metkey);
	if (prev != null)
	    assign = prev + " " + assign;
	setProperty(metkey, assign);

        // Save the time
	lastTime = getTime();

	// Make and show the window
	/*window = new JFrame(getPlugInName());
	window.getContentPane().add(driverpanel = new DriverPanel(this));
	window.setSize(200,100);
	window.setVisible(true);
	window.pack();
	
	// Listen
	window.addWindowListener(new WindowAdapter() {
		public void windowClosing(WindowEvent e) {
		    setStop();
		}
	    });    

	driverpanel.setTargetText(targSpeed);
	*/


    }

    /**
     */
    public void end() throws RemoteException {
	
        // Deal with the window
        if (window != null)
            window.dispose();
	
        super.end();
    }

    private ArrayList createSensors(int width, int height, int numSens){
        int placemodel = 0;
        float numx = 0, numy = 0, xdelta = 0, ydelta = 0;

        // Find the placement model
        if (placement.equalsIgnoreCase("random")) {
            placemodel = 0;
        } else if (placement.equalsIgnoreCase("ordered")) {
            placemodel = 1;
            numx = (int)Math.ceil(Math.pow(numSens, (float)width / (float)(width + height)));
            numy = (int)Math.ceil(Math.pow(numSens, (float)height / (float)(width + height)));
            xdelta = (float)width / numx;
            ydelta = (float)height / numy;
        } else {
            log.log("Unknown placement model " + placement, Log.LOG_INFO);
        }

        // Place them
	ArrayList sens = new ArrayList();
        float x = 0, y = -ydelta / 2;
	for(int i = 0;i<numSens;i++){
	    String name = "SA" + i;

            switch (placemodel) {
            case 1: // ordered
                if (i % numx == 0) {
                    x = xdelta / 2;
                    y += ydelta;
                } else {
                    x += xdelta;
                }
                break;
            case 0: // random
            default:
                x = rnd.nextFloat() * width;
                y = rnd.nextFloat() * height;
            }

	    SensorModel sensMdl = new SensorModel(this, name, x, y, sensRange);
	    sens.add(sensMdl);
	}
	Collections.sort(sens);
	return sens;
    }

    private ArrayList createTargets(int width, int height, int numTargs){
	ArrayList targs = new ArrayList();
	for(int i = 0;i<numTargs;i++){
	    String name = "M" + i;
	    TargetModel targMdl = new TargetModel(this, name, 
						  rnd.nextFloat() * width, 
						  rnd.nextFloat() * height, 
						  targSpeed);
	    
	    targMdl.setRoomWidth(width);
	    targMdl.setRoomHeight(height);
	    targMdl.setSensors(findSensors(targMdl, sensors));

	    targs.add(targMdl);
	}
	return targs;
    }

    private ArrayList findSensors(TargetModel targ, ArrayList sensors){
	//find the right sensors for the target
	//the sensors list should be sorted by object model comparator

	ArrayList ret = new ArrayList();
	Point2D.Float targLoc = targ.getPoint(false);
	int lowX = (int) (targLoc.getX() - sensRange);
	int highX = (int) (targLoc.getX() + sensRange);

	for(int i = 0;i<sensors.size();i++){
	    SensorModel sens = (SensorModel) sensors.get(i);
	    Point2D.Float loc = sens.getPoint(false);
	    if(loc.getX() < lowX) continue;
	    if(loc.getX() >= highX) break;
	    SensorData sensData = new SensorData(sens.getName(), 
						 (float) loc.distance(targLoc));
	    if(loc.distance(targLoc) <sensRange && !ret.contains(sensData)){
		ret.add(sensData);
	    }
	}
	//sort the list
	Collections.sort(ret);

	return ret;
    }

    /**
     * Main Function
     */
    public static void main(String[] args) {
	try {
	    FarmDriver driver = new FarmDriver(args);
	    driver.run();
	} catch (RemoteException e) {
            System.err.println("Exception: " + e.getMessage());
            e.printStackTrace();
        }
    }    

} // FarmDriver

/**
 * The Driver window
 */
class DriverPanel extends JPanel {
    JTextField text;

    FarmDriver driver;

    public DriverPanel(FarmDriver d) {
        super();
        
        // Save the reference
        driver = d;;

        // Set up the panel
        text = new JTextField(6);
        add(text);
	JButton set = new JButton("Set");
        add(set);

	JButton pause = new JButton(driver.isPaused() ? "Unpause" : "  Pause  ");
        add(pause);

        // Listen to their stories
        set.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    String str = text.getText();
                    driver.resetTargetSpeed(Float.parseFloat(str));
                }
            });
    
        pause.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    driver.setPaused(!driver.isPaused());
                    if (driver.isPaused()) {
                        ((JButton)e.getSource()).setText("Unpause");
                    } else {
                        ((JButton)e.getSource()).setText("  Pause  ");
                    }
                }
            });
    }

    public void setTargetText(float t) {
        text.setText(String.valueOf(t));
    }
}
