package farm.generic.plugins.drivers;

import java.rmi.*;
import java.rmi.server.*;
import farm.simulator.*;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import java.util.*;

/**
 * This plugin is used to "drive" the time aspect of the simulation.  Unlike
 * most simulation environments, the farm simulator itself has no notion of time
 * (other than the fact that a property P_TIME exists which contains the
 * time).  Thus, an outside entity is needed to implement a notion of time.
 * The time driver does this, by incrementing the global P_TIME property by
 * some amount each time it is pulsed.
 * <p>
 * Properties:<dl>
 * <dt><tt>-delta [integer]</tt>
 * <dd>The integer amount by which time is incremented each pulse.
 *
 * <dt><tt>-paused</tt>
 * <dd>If true, time is paused at startup
 * 
 * <dt><tt>-gui</tt>
 * <dd>If set, the TimeDriver's GUI will be displayed, which allows the operator
 * to manually pause and end the simulation.
 *
 * <dt><tt>-end [integer]</tt>
 * <dd>If included, TimeDriver will end the simulation at the specified time
 *
 * <dt><tt>-adjusttime</tt>
 * <dd>If set, the driver will attempt to make the overall running time
 * conform to the specified time by sleeping when it notices pulses
 * are too short.
 * </dl>
 */
public class TimeDriver extends FarmPlugIn {

    // Constants
    public static final String CLA_TIMEDELTA = "delta";
    public static final String CLA_TIMEPAUSED = "paused";
    public static final String CLA_TIMEGUI = "gui";
    public static final String CLA_ENDTIME = "end";
    public static final String CLA_ADJUSTTIME = "adjusttime";
    public static int DEFAULT_TIME = 100;

    // Local information
    protected int delta = 0;
    protected boolean paused = false;
    protected boolean usegui = false;
    protected boolean adjusttime = false;
    protected int endtime = -1;
    protected JFrame window;
    protected TimePanel timepanel;
    protected long lastpulse = System.currentTimeMillis();
    protected long averagepulse = Long.MIN_VALUE;

    /**
     * Constructor
     * @param address The address of the remote simulator to eonnect to
     */
    public TimeDriver(String []args) throws RemoteException {
        super(args);

        delta = DEFAULT_TIME;

        // Command line args
        setTimeDelta(Integer.parseInt(getCommandLineArgument(args, CLA_TIMEDELTA, String.valueOf(DEFAULT_TIME))));
        setPlugInName(getCommandLineArgument(args, CLA_NAME, "Father Time"));
        setEndTime(Integer.parseInt(getCommandLineArgument(args, CLA_ENDTIME, String.valueOf(endtime))));
        paused = (getCommandLineArgument(args, CLA_TIMEPAUSED, null) != null);
        usegui = (getCommandLineArgument(args, CLA_TIMEGUI, null) != null);
        adjusttime = (getCommandLineArgument(args, CLA_ADJUSTTIME, null) != null);
    }

    /**
     * Sets the time delta
     */
    public void setTimeDelta(int i) {
        delta = i;
    }

    /**
     * Gets the time delta
     */
    public int getTimeDelta() {
        return delta;
    }

    /**
     * Sets the paused status
     */
    public void setPaused(boolean p) {
        paused = p;
    }
    public boolean isPaused() {
        return paused;
    }

    /**
     * Terminates the run
     */
    public void stopSimulation() {
        try {
            getSimulator().setStop();
        } catch (RemoteException e) {
            System.err.println("Exception: " + e.getMessage());
            e.printStackTrace();
        }
    }

    /**
     * Sets the time when the simulation will be terminated
     * by the time driver.
     * @param t The time when the simulation will end.  Set to
     *   zero or less to run forever.
     */
    public void setEndTime(int t) {

        endtime = t;
    }

    /**
     * Sets the current time
     */
    public void setTime(int t) {

        setProperty(Simulator.P_TIME, new Integer(t));
        if (timepanel != null) {
            timepanel.setTimeText(t);
        }
    }

    /**
     */
    public void begin() throws RemoteException{
        super.begin();

        lastpulse = System.currentTimeMillis();

        if (usegui) {
            // Make and show the window
            window = new JFrame(getPlugInName());
            window.getContentPane().add(timepanel = new TimePanel(this));
            window.setSize(200,100);
            window.setVisible(true);
            window.pack();

            // Listen
            window.addWindowListener(new WindowAdapter() {
                    public void windowClosing(WindowEvent e) {
                        setStop();
                    }
                });

            // Thread
            Thread realtime = new Thread() {
                    public void run() {
                        while (!isStopping()) {
                            timepanel.setRealTimeText(System.currentTimeMillis() - lastpulse);
                            mysleep(100);
                        }
                    }
                };
            realtime.start();
        }
    }

    /**
     * This is called at the beginning of a pulse, before the
     * agents are started.
     */
    public void pulseBegin() throws RemoteException {
	while(isPaused()){
	    mysleep(20);
	}
        if (!isPaused()) {
            // Delay if we should
            if (adjusttime) {
		    if ((lastpulse + delta) > System.currentTimeMillis())
			mysleep((int) ((lastpulse + delta)- System.currentTimeMillis()));
	    }

            // Set the current time
            Integer time = (Integer)getProperty(Simulator.P_TIME);
            setTime((time == null) ? 0 : time.intValue() + delta);

            // Compute the average time
            if (averagepulse == Long.MIN_VALUE)
                averagepulse = System.currentTimeMillis() - lastpulse;
            else
                averagepulse = averagepulse + (long)(0.4f * (float)((System.currentTimeMillis() - lastpulse) - averagepulse));
            timepanel.setAverageTimeText(averagepulse);
	    lastpulse = System.currentTimeMillis();
        }
    }

    /**
     */
    public void pulseEnd() throws RemoteException {

        if (endtime > 0) {
            if (!isPaused()) {
                Integer time = (Integer)getProperty(Simulator.P_TIME);
                if (time.intValue() >= endtime) {
                    stopSimulation();
                }
            }
        }
    }

    /**
     */
    public void end() throws RemoteException {

        // Deal with the window
        if (window != null)
            window.dispose();

        super.end();
    }

    /**
     * The main function
     */
    public static void main(String[] args) {

        try {
            FarmPlugIn plugin = new TimeDriver(args);
            plugin.run();
        } catch (RemoteException e) {
            System.err.println("Exception: " + e.getMessage());
            e.printStackTrace();
        }
    }
}

/**
 * The time window
 */
class TimePanel extends JPanel {
    JTextField timetext;
    JTextField deltatext;
    JLabel realtext;
    JLabel avgtext;

    TimeDriver time;

    public TimePanel(TimeDriver t) {
        super();
        
        // Save the reference
        time = t;

        // Pick a layout
        setLayout(new GridLayout(0, 1));

        // Set up the panels
        JPanel panel = new JPanel();
        JButton pause = new JButton(time.isPaused() ? "Unpause" : "  Pause  ");
        panel.add(pause);
        JButton kill = new JButton("End");
        panel.add(kill);
        add(panel);

        panel = new JPanel();
        panel.setLayout(new FlowLayout(FlowLayout.LEFT));
        panel.add(new JLabel("Time:"));
        timetext = new JTextField(6);
        panel.add(timetext);
        JButton settime = new JButton("Set");
        panel.add(settime);
        add(panel);

        panel = new JPanel();
        panel.setLayout(new FlowLayout(FlowLayout.LEFT));
        panel.add(new JLabel("Delta:"));
        deltatext = new JTextField(6);
        deltatext.setText(String.valueOf(time.getTimeDelta()));
        panel.add(deltatext);
        JButton setdelta = new JButton("Set");
        panel.add(setdelta);
        add(panel);

        panel = new JPanel();
        panel.setLayout(new FlowLayout(FlowLayout.LEFT));
        panel.add(new JLabel("Real Pulse Time:"));
        realtext = new JLabel("       ");
        panel.add(realtext);
        add(panel);

        panel = new JPanel();
        panel.setLayout(new FlowLayout(FlowLayout.LEFT));
        panel.add(new JLabel("Average Pulse Time:"));
        avgtext = new JLabel("       ");
        panel.add(avgtext);
        add(panel);

        // Listen to their stories
        settime.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    String str = timetext.getText();
                    time.setTime(Integer.parseInt(str));
                }
            });
        setdelta.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    String str = deltatext.getText();
                    time.setTimeDelta(Integer.parseInt(str));
                }
            });
        pause.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    time.setPaused(!time.isPaused());
                    if (time.isPaused()) {
                        ((JButton)e.getSource()).setText("Unpause");
                    } else {
                        ((JButton)e.getSource()).setText("  Pause  ");
                    }
                }
            });
        kill.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    time.stopSimulation();
                }
            });
    }

    public void setTimeText(long t) {

        timetext.setText(String.valueOf(t));
    }

    public void setRealTimeText(long t) {

        realtext.setText(String.valueOf(t));
    }

    public void setAverageTimeText(long t) {

        avgtext.setText(String.valueOf(t));
    }
}
