package farm.graphcolor.agent;

import farm.simulator.*;
import farm.metagent.*;
import farm.graphcolor.plugins.drivers.*;

import java.util.*;
import utilities.Log;

/**
 * Implements Asynchronous Back Tracking Search
 */
public class ABTAgent extends Agent {
    NodeModel model;

    Hashtable agent_view = new Hashtable();
    Collection nogood_list = new ArrayList();
    Collection neighbors = new TreeSet();

    /**
     * Constructor
     * @param metagent A handle to the metagent this agent belongs to
     * @param m The NodeModel assigned to this agent
     */
    public ABTAgent(FarmMetAgent metagent, NodeModel m) {
	super(metagent, m.getName());

        model = m;

        // Initialize
        for (Iterator i = model.getNeighbors(); i.hasNext(); ) {
	    String neigh = (String) i.next();
            neighbors.add(neigh);
	}
	setProperty(model.getName() + ":links" , new ArrayList(neighbors));
	sendOK();
     }

    /**
     * Looks at incoming messages
     */
    public void pulse() {
	super.pulse();

        // Respond to messages
        if (hasIncomingMessage()) {
            for (Message msg = getFirstIncomingMessage(); msg != null; msg = getFirstIncomingMessage()) {
                checkRun();

                if (msg.getData().startsWith("ok? ")) {
                    ABTAgentView v = new ABTAgentView(msg.getData().substring(4));
                    agent_view.put(v.getName(), v);
		    checkAgentView();

                } else if (msg.getData().startsWith("nogood ")) {
                    Collection set = parseAllAgentViews(msg.getData().substring(7));
                    nogood_list.add(set);
                    for (Iterator i = set.iterator(); i.hasNext(); ) {
                        ABTAgentView av = (ABTAgentView)i.next();
                        if (!neighbors.contains(av.getName()) && !av.getName().equals(model.getName())) {
                            log.log("Adding new neighbor " + av.getName(), Log.LOG_INFO);
                            neighbors.add(av.getName());
			    sendMessage(new Message(getAgentName(), av.getName(), "link"));
                            agent_view.put(av.getName(), av);
			    setProperty(model.getName() + ":links" , new ArrayList(neighbors));
                        }
                    }
		    int oldColor = model.getColor();
		    checkAgentView();
		    if(oldColor == model.getColor()){
			ABTAgentView av = new ABTAgentView(model);
			sendMessage(new Message(getAgentName(), msg.getSource(), "ok? " + av.toString()));
		    }
                } else if (msg.getData().startsWith("link")) {
		    neighbors.add(msg.getSource());
		    ABTAgentView av = new ABTAgentView(model);
		    sendMessage(new Message(getAgentName(), msg.getSource(), "ok? " + av.toString()));
		}
            }
        }
        endRun();
    }

    /**
     * Checks if the agent view is consistant
     */
    public void checkAgentView() {

        if (!isViewConsistent(agent_view, model.getColor())) {
            // Pick a new color
            int bestcolor = chooseBestColor();
            int mycolor = model.getColor();

            // Is there a valid color?
            if (bestcolor == -1) {
                backtrack();

            // Set the color
            } else {
                log.log(getAgentName() + " changing color from " + mycolor + " to " + bestcolor, Log.LOG_INFO);
                model.setColor(bestcolor);
		sendOK();
            }
        }
    }

    /**
     * Backtracks in the search
     */
    public void backtrack() {

        // Get the current agent view
        Collection nogoods = new TreeSet();

        for (Iterator i = agent_view.values().iterator(); i.hasNext(); ) {
            ABTAgentView av = (ABTAgentView)i.next();
            nogoods.add(av.clone());
        }

        // The nogood string
        String ng = generateAgentView(nogoods);

	// Send it to everyone
	log.log(getAgentName() + " is nogood: " + ng, Log.LOG_INFO);

	ABTAgentView lowman = getLowestPriority();

	sendMessage(new Message(getAgentName(), lowman.getName(), "nogood " + ng));
	agent_view.remove(lowman.getName());
	checkAgentView();
    }

    /**
     * Determines if the current setting is consistent or not
     */
    public boolean isViewConsistent(Hashtable view, int color) {

        // No color is bad
        if (color == -1) return false;

        // Check neighbors
        for (Iterator i = model.getNeighbors(); i.hasNext();) {
            String n = (String)i.next();
            ABTAgentView v = (ABTAgentView)view.get(n);
            if (v != null && isHigherPriority(n)) {
                if (color == v.getColor()) return false;
            }
        }

        // Check nogoods
        return (!isAgentViewNoGood(view, color));
    }

    /**
     * Selects the best color, based on local knowledge
     */
    public int chooseBestColor() {
        int colors = model.getPossibleColors();
        int best = -1;

        // Find the best color
        for (int i = 0; i < colors; i++) {
	    if (isViewConsistent(agent_view, i)) {
		best = i;
		break;
	    }
        }

        log.log("Best color is now " + best, Log.LOG_INFO);
        return best;
    }

    /**
     * Get priority of the lowest priority agent in the agent view
     */
    public ABTAgentView getLowestPriority() {
        int m = Integer.MAX_VALUE;
	ABTAgentView best = null;

        for (Enumeration e = agent_view.elements(); e.hasMoreElements(); ) {
            ABTAgentView agent =  (ABTAgentView)e.nextElement();
	    int i = agent.getPriority();
            if (i < m) {
                m = i;
		best = agent;
            }
        }

        return best;
    }

    /**
     * Determines if an agent is more important than me
     */
    public boolean isHigherPriority(String n) {
	int p = Integer.parseInt(n.substring(2));
	int m = Integer.parseInt(model.getName().substring(2));
	return (p > m);
    }

    /**
     * Generates an agent view string
     */
    public String generateAgentView(Collection view) {
        StringBuffer buf = new StringBuffer();

        for (Iterator i = view.iterator(); i.hasNext(); ) {
            ABTAgentView av = (ABTAgentView)i.next();
            buf.append(av.toString());
            if (i.hasNext()) {
                buf.append(",");
            }
        }

        return buf.toString();
    }

    /**
     * Construct's a string of the agents' views
     */
    public Collection parseAllAgentViews(String s) {
        Collection set = new TreeSet();
        StringTokenizer tok = new StringTokenizer(s, ",");

        while (tok.hasMoreTokens()) {
            set.add(new ABTAgentView(tok.nextToken()));
        }

        return set;
    }

    /**
     * Given a new color for myself, determines if the resulting
     * local agent view contains any nogoods.
     */
    public boolean isAgentViewNoGood(Hashtable view, int c) {
        ABTAgentView myview = new ABTAgentView(model.getName(), c);
        view.put(myview.getName(), myview);
        boolean good = true;

        for (Iterator i = nogood_list.iterator(); i.hasNext(); ) {
            Collection set = (Collection)i.next();
            if (view.values().containsAll(set)) {
                good = false;
                break;
            }
        }

        view.remove(myview.getName());

        //log.log("Value " + c + " is good? " + good, Log.LOG_INFO);

        return !good;
    }

    private void sendOK(){
    	ABTAgentView av = new ABTAgentView(model);
        for (Iterator i = neighbors.iterator(); i.hasNext();) {
	    String neigh = (String) i.next();
	    if(!isHigherPriority(neigh)){
		sendMessage(new Message(getAgentName(), neigh, "ok? " + av.toString()));
	    }
	}
    }

}

class ABTAgentView implements Cloneable, Comparable {
    String name;
    int color;
    int priority;

    public ABTAgentView(String n, int c) {
        name = n;
        color = c;
	priority = Integer.parseInt(name.substring(2));
    }
    public ABTAgentView(String s) {
        StringTokenizer tok = new StringTokenizer(s);
        name = tok.nextToken();
        color = Integer.parseInt(tok.nextToken());
        priority = Integer.parseInt(name.substring(2));
    }
    public ABTAgentView(NodeModel n) {
        this(n.getName(), n.getColor());
    }
    public ABTAgentView(ABTAgentView v) {
        this(v.getName(), v.getColor());        
    }

    public String getName() { return name; }
    public int getColor() { return color; }
    public int getPriority() { return priority; }

    public Object clone() {
        return new ABTAgentView(this);
    }

    public boolean equals(Object o) {
        ABTAgentView v = (ABTAgentView)o;
        if (v.getColor() == color &&
             v.getName().equals(name))
            return true;
        return false;
    }

    public int compareTo(Object o) {
        ABTAgentView v = (ABTAgentView)o;

        if (equals(v)) return 0;

        return (priority > v.getPriority()) ? 1 : -1;
    }

    public String toString() {
        return name + " " + color;
    }
}
