package farm.graphcolor.agent;

import farm.simulator.*;
import farm.metagent.*;
import farm.graphcolor.plugins.drivers.*;
import farm.generic.plugins.drivers.*;
import java.util.*;
import utilities.Log;

/**
 * Implements Asynchronous Weak Commitment Search
 */
public class AWCAgent extends Agent {
    NodeModel model;

    Hashtable agent_view = new Hashtable();
    Collection nogood_list = new ArrayList();
    Collection nogood_sent = new ArrayList();
    Collection neighbors = new TreeSet();

    /**
     * Constructor
     * @param metagent A handle to the metagent this agent belongs to
     * @param m The NodeModel assigned to this agent
     */
    public AWCAgent(FarmMetAgent metagent, NodeModel m) {
	super(metagent, m.getName());

        model = m;
	//log.log("My given model: " + m.toString(), Log.LOG_INFO);
	log.setLevel(Log.LOG_ERR);

	if(model.numNeighbors() > 0){
	    setProperty(model.getName() + ":done", new Boolean(false));
	}
	else{
	    setProperty(model.getName() + ":done", new Boolean(true));
	}

        // Initialize
        for (Iterator i = model.getNeighbors(); i.hasNext();) {
	    String neigh = (String) i.next();
            neighbors.add(neigh);

	    //log.log("Adding neighbor " + neigh,Log.LOG_INFO);
	    //add the nogoods for the problems constraints
	    int possColors = model.getPossibleColors();
	    for(int col = 0;col < possColors; col++){
	        Nogood ng = new Nogood();
		ng.add(new NGPair(m.getName(), "" + col));
		ng.add(new NGPair(neigh, "" + col));
		nogood_list.add(ng);
		//log.log("Added nogood " + ng.toString(),Log.LOG_INFO);
	    }
        }

	//setProperty(model.getName() + ":links" , new ArrayList(neighbors));

	sendOK();
 	//checkAgentView();
     }

    /**
     * Looks at incoming messages
     */
    public void pulse() {
	super.pulse();

        // Respond to messages
        if (hasIncomingMessage()) {
            for (Message msg = getFirstIncomingMessage(); msg != null; msg = getFirstIncomingMessage()) {
                //checkRun();

                if (msg.getData().startsWith("ok? ")) {
                    AWCAgentView v = new AWCAgentView(msg.getData().substring(4));
                    agent_view.put(v.getName(), v);
		    checkAgentView();
		    
                } else if (msg.getData().startsWith("nogood ")) {
                    Nogood ng = new Nogood();
		    ng.fromString(msg.getData().substring(7));
                    nogood_list.add(ng);
                    for (Iterator i = ng.iterator(); i.hasNext(); ) {
			//checkRun();
                        NGPair ngp = (NGPair)i.next();
                        if (!neighbors.contains(ngp.getName()) && !ngp.getName().equals(model.getName())) {
                            log.log("Adding new neighbor " + ngp.getName(), Log.LOG_INFO);
                            neighbors.add(ngp.getName());
                            sendMessage(new Message(getAgentName(), ngp.getName(), "link"));
			    setProperty(model.getName() + ":links" , new ArrayList(neighbors));
                        }
                    }
		    int old_value = model.getColor();
		    checkAgentView();
		    if(old_value == model.getColor()){
			sendMessage(new Message(getAgentName(), msg.getSource(), "ok? " + new AWCAgentView(model).toString()));
		    }
                } else if (msg.getData().startsWith("link")) {
		    neighbors.add(msg.getSource());
		    AWCAgentView av = new AWCAgentView(model);
		    sendMessage(new Message(getAgentName(), msg.getSource(), "ok? " + av.toString()));
		}
            }
        }
        endRun();
    }

    /**
     * Checks if the agent view is consistant
     */
    public void checkAgentView() {

        if (!isViewConsistent(agent_view, model.getColor())) {
 	    setProperty(model.getName() + ":done", new Boolean(false));
           // Pick a new color
            int bestcolor = chooseBestColor();
            int mycolor = model.getColor();
	    //checkRun();

            // Is there a valid color?
            if (bestcolor == -1) {
                backtrack();

            // Set the color
            } else if (bestcolor != mycolor) {
                log.log(getAgentName() + " changing color from " + mycolor + " to " + bestcolor, Log.LOG_INFO);
                model.setColor(bestcolor);
		setProperty(model.getName() + ":done", new Boolean(true));
		sendOK();
            }
	}
	else{
	    setProperty(model.getName() + ":done", new Boolean(true));
	}
    }

    /**
     * Backtracks in the search
     */
    public void backtrack() {
        // Generate the nogood from the current agent view
        Nogood nogood = generateNogood(agent_view.values());

	if(nogood.isNull()){
	    //there is no solution to the problem
	   log.log("!!!!!!Problem is unsolvable", Log.LOG_INFO);
	   setProperty("Solved", new Boolean(false));
	   setProperty(StopDriver.STOPSIMULATION, new Boolean(true));
	   return;
	}

        // The nogood string
        String ng = nogood.toString();

        // If we haven't seen this before
        if (!nogood_sent.contains(nogood)) {
            // Save it
            nogood_sent.add(nogood);
	    //checkRun();

            // Send it to everyone
            log.log(getAgentName() + " is nogood: " + ng, Log.LOG_INFO);
            for (Iterator i = nogood.iterator(); i.hasNext(); ) {
                NGPair ngp = (NGPair)i.next();
                sendMessage(new Message(getAgentName(), ngp.getName(), "nogood " + ng));
            }
	    //checkRun();

            // Pick new color
            model.setPriority(getHighestPriority() + 1);
            int mycolor = model.getColor();
            int bestcolor = chooseBestColor();

            log.log(getAgentName() + " changing color from " + mycolor + " to " + bestcolor, Log.LOG_INFO);
            model.setColor(bestcolor);
	    sendOK();
        } else {
            log.log("I've already sent this nogood: " + ng, Log.LOG_INFO);
        }
    }

    /**
     * Determines if the current setting is consistent or not
     * @return true is consistent false otherwise
     */
    public boolean isViewConsistent(Hashtable view, int color) {

        // No color is bad
        if (color == -1) return false;

        AWCAgentView myview = new AWCAgentView(model.getName(), color, model.getPriority());
        view.put(myview.getName(), myview);
        boolean good = true;
	//log.log("Current view " + view.toString(), Log.LOG_INFO);
        for (Iterator i = nogood_list.iterator(); i.hasNext(); ) {
            Nogood ng = (Nogood) i.next();
	    //checkRun();
	    //log.log("Checking nogood " + ng.toString(),Log.LOG_INFO);
	    if(ng.isHigherPriority(model.getName(), view, model)){
		//log.log("nogood is higher priority", Log.LOG_INFO);
		//its a higher priority nogood
		if(ng.isViolated(view)){
		    log.log("nogood " + ng.toString() + " is violated", Log.LOG_INFO);
		    //checkRun();
		    good = false;
		    break;
		}
            }
        }
	
        view.remove(myview.getName());
        //log.log("Check nogood Value " + color + " is good? " + good, Log.LOG_INFO);
        return good;	
    }

    /**
     * Selects the best color, based on local knowledge
     * @param consistent If true, the consistancy of the assignment is checked
     */
    public int chooseBestColor() {
	int possColors = model.getPossibleColors();
        int colors[] = new int[possColors];
        int mycolor = model.getColor();
        int best = -1;

        AWCAgentView myview = new AWCAgentView(model.getName(), mycolor, model.getPriority());
        agent_view.put(myview.getName(), myview);

        // Find the best color
        for (int color = 0; color < possColors; color++) {
	    //checkRun();

	    //does this violate a higher priority nogood
	    if(!isViewConsistent(agent_view, color)){
		colors[color] = -1;
		continue;
	    }
	    myview.setColor(color);
	    agent_view.put(myview.getName(), myview);
	    //this color just violates lower priority nogoods, find out
	    //how many
	    for (Iterator i = nogood_list.iterator(); i.hasNext(); ) {
                //checkRun();
		Nogood ng = (Nogood) i.next();
		//log.log("ng: " + ng.toString() + " view: " + agent_view.toString(), Log.LOG_INFO);
		if(!ng.isHigherPriority(model.getName(), agent_view, model) && 
		   ng.isViolated(agent_view)){
			colors[color] ++;
		}
	    }
	}	

        // Find the best color
        for (int color = 0; color < colors.length; color++) {
	    if(colors[color] == -1) continue;  //violates a higher priority

            // If one is clearly better, choose it
            if (best == -1 || colors[color] < colors[best]) {
		best = color;
            }
        }
        agent_view.remove(myview.getName());
	//checkRun();

        log.log("Best color is now " + best, Log.LOG_INFO);
        return best;
    }

    /**
     * Get priority of the highest priority agent we know about
     */
    public int getHighestPriority() {
        int m = 0;

        for (Enumeration e = agent_view.elements(); e.hasMoreElements(); ) {
            int i = ((AWCAgentView)e.nextElement()).getPriority();
            if (i > m) {
                m = i;
            }
        }
        return m;
    }

    /**
     * Determines if an agent is more important than me
     */
    public boolean isHigherPriority(String n) {
        AWCAgentView v = (AWCAgentView)agent_view.get(n);
	AWCAgentView myself = new AWCAgentView(model);

	if(myself.compareTo(v) > 0) return true;
        return false;
    }

    /**
     * Generates an agent view string
     */
    public Nogood generateNogood(Collection view) {
	int possColors = model.getPossibleColors();
	Nogood ng = new Nogood();
	ArrayList violated = new ArrayList(possColors); 
        AWCAgentView myview = new AWCAgentView(model.getName(), model.getColor(), model.getPriority());
        agent_view.put(myview.getName(), myview);
	String myname = model.getName();


	//for each color, idenitfy the highest priority violated nogood
        for (int color = 0; color < possColors; color++) {
	    Nogood thisColorBest = null;
	    myview.setColor(color);
	    agent_view.put(myview.getName(), myview);
	    //checkRun();

	    for (Iterator i = nogood_list.iterator(); i.hasNext(); ) {
		Nogood testNG = (Nogood) i.next();
		//checkRun();
		//log.log("testNG: " + testNG.toString() + " bestNG " + ((thisColorBest == null)?"NULL":thisColorBest.toString()) +" view: " + agent_view.toString(), Log.LOG_INFO);
		if(testNG.isHigherPriority(model.getName(), agent_view, model) && testNG.isHigherPriority(thisColorBest, agent_view, model) && 
		   testNG.isViolated(agent_view)){
		    //log.log("testNG is better", Log.LOG_INFO);
		    thisColorBest = testNG;
		}
	    }
	    log.log("Highest priority nogood for color " + color + " is " + thisColorBest.toString(),Log.LOG_INFO);
	    violated.add(thisColorBest);
	}

	//resolve them
	for(Iterator viol = violated.iterator();viol.hasNext();){
	    Nogood thisNG = (Nogood) viol.next();
	    for(Iterator pairs = thisNG.iterator();pairs.hasNext();){
                //checkRun();
		NGPair NGP = (NGPair) pairs.next();
		if(!NGP.getName().equals(myname))
		    ng.add(NGP);
	    }
	}



	/*Non resolvent method
	for (Iterator i = view.iterator(); i.hasNext(); ) {
	    AWCAgentView av = (AWCAgentView)i.next();
	    if(isHigherPriority(av.getName())){
		NGPair ngp = new NGPair(av.getName(), av.getValue());
		ng.add(ngp);
	    }
	}
	*/

	return ng;
    }

    private void sendOK(){
    	AWCAgentView av = new AWCAgentView(model);
        for (Iterator i = neighbors.iterator(); i.hasNext();) {
	    String neigh = (String) i.next();
	    sendMessage(new Message(getAgentName(), neigh, "ok? " + av.toString()));
	}
    }

}

class AWCAgentView implements Cloneable, Comparable {
    String name;
    int color;
    int priority;

    public AWCAgentView(String n, int c, int p) {
        name = n;
        color = c;
        priority = p;
    }
    public AWCAgentView(String s) {
        StringTokenizer tok = new StringTokenizer(s);
        name = tok.nextToken();
        color = Integer.parseInt(tok.nextToken());
        priority = Integer.parseInt(tok.nextToken());
    }
    public AWCAgentView(NodeModel n) {
        this(n.getName(), n.getColor(), n.getPriority());
    }
    public AWCAgentView(AWCAgentView v) {
        this(v.getName(), v.getColor(), v.getPriority());        
    }

    public String getName() { return name; }
    public int getColor() { return color; }
    public int getPriority() { return priority; }
    public String getValue() {return "" + color;}
    public void setColor(int col){color = col;}
    public Object clone() {
        return new AWCAgentView(this);
    }

    public boolean equals(Object o) {
        AWCAgentView v = (AWCAgentView)o;
        if (v.getName().equals(name))
            return true;
        return false;
    }

    
    public int compareTo(Object o) {
        AWCAgentView v = (AWCAgentView)o;

        if (equals(v)) return 0;

        if (v.getPriority() == priority) {
	    int i1 = Integer.parseInt(name.substring(2));
	    int i2 = Integer.parseInt(v.getName().substring(2));
            return (i1 > i2) ? -1 : 1;
        }
        return (priority > v.getPriority()) ? -1 : 1;
    }

    public String toString() {
        return name + " " + color + " " + priority;
    }
}

class NGPair implements Cloneable {

    String name;
    String value;

    public NGPair(){
    }

    public NGPair(String n, String v){
	setName(n);
	setValue(v);
    }

    public NGPair(NGPair ngp){
	setName(ngp.getName());
	setValue(ngp.getValue());
    }

    public void setName(String n){name = n;}
    public String getName(){return name;}
    public void setValue(String v){value = v;}
    public String getValue(){return value;}

    public Object clone() {
        return new NGPair(this);
    }

    public boolean equals(Object o) {
        NGPair ngp = (NGPair)o;
        if (ngp.getName().equals(name) &&
            ngp.getValue().equals(value))
            return true;
        return false;
    }

    
    public String toString() {
        return "(" + name + " " + value + ")";
    }

    public void fromString(String val){
	StringTokenizer tok = new StringTokenizer(val,"() ");
	name = tok.nextToken();
	value = tok.nextToken();
    }
}

class Nogood extends ArrayList{
    

    public void add(NGPair ngp){
	if(!super.contains(ngp))
	    super.add(ngp);
    }


    /**
     * Tests to see if this nogood is higher priority than the agent named
     * based on the current agent view
     * @return true if nogood is higher priority
     */
    public boolean isHigherPriority(String who, Hashtable agent_view, NodeModel model){	
	boolean higher = true;
	AWCAgentView avWho = (AWCAgentView) agent_view.get(who);

	for(Iterator i = this.iterator();i.hasNext();){
	    NGPair ngp = (NGPair) i.next();
	    if(ngp.getName().equals(model.getName())) continue;
	    if(agent_view.containsKey(ngp.getName())){
		AWCAgentView av = (AWCAgentView) agent_view.get(ngp.getName());
		if(avWho.compareTo(av) < 0){
		    higher = false;
		    break;
		}
	    }
	    else{
		higher = false;
		break;
	    }
	}
	return higher;
    }

    /**
     * Tests to see if one nogood is higher priority than another based on the
     * current agent view
     * @returns true if this is better, false otherwise
     */
    public boolean isHigherPriority(Nogood ng, Hashtable agent_view, NodeModel model){

	if(ng==null) return true;

	//smaller ones are always better
	if(ng.size() < this.size()) return false;
	//bigger ones are always worse
	if(ng.size() > this.size()) return true;
	
	//equal sized ones are based on the highest priority in each
	for(Iterator pairs = ng.iterator();pairs.hasNext();){
	    NGPair NGP = (NGPair) pairs.next();
	    if(!NGP.getName().equals(model.getName()) && 
	       !this.isHigherPriority(NGP.getName(), agent_view, model))
		return false;
	}
	return true;
    }

    /**
     * Tests to see is a nogood is violated based on the current view
     * @param agent_view the current view
     * @return true is violated false otherwise
     */
    public boolean isViolated(Hashtable agent_view){
	boolean violated = true;
	for(Iterator i = this.iterator();i.hasNext();){
	    NGPair ngp = (NGPair) i.next();
	    if(!agent_view.containsKey(ngp.getName())){
		violated = false;
		break;
	    }
	    else{
		AWCAgentView av = (AWCAgentView) agent_view.get(ngp.getName());
		if(!av.getValue().equals(ngp.getValue())){
		    violated = false;
		    break;
		}
	    }
	}
	return violated;
    }
    
    /**
     * Stringify
     */
    public String toString(){
	StringBuffer sBuff = new StringBuffer();
	for(Iterator i = this.iterator();i.hasNext();){
	    sBuff.append(((NGPair)i.next()).toString());
	    if(i.hasNext()) sBuff.append(":");
	}
	return sBuff.toString();
    }


    /**
     * Destringify
     */
    public void fromString(String s){
	StringTokenizer tok = new StringTokenizer(s,":");
	while(tok.hasMoreTokens()){
	    NGPair newNGP = new NGPair();
	    newNGP.fromString(tok.nextToken());
	    this.add(newNGP);
	}
    }

    public boolean equals(Object oNG){
	
	Nogood NG = (Nogood) oNG;
	if(NG.size() != this.size()) return false;
	for(Iterator pairs = this.iterator();pairs.hasNext();){
	    if(!NG.contains(pairs.next())) return false;
	}
	return true;
    }
    public boolean isNull(){
	return (this.size() ==0);
    }

}
