/**
 */

package farm.metagent;

import farm.simulator.Simulator;

import java.util.*;

import utilities.Log;

/**
 * This is the base class for farm agents.  A metagent will typically create
 * one or more of these when it starts, after which they (should) exist in
 * a sort of local vacuum to better model actually distributed runtime conditions.
 * The bulk of agent activity is typically implemented from the pulse function,
 * which is called continuously during the agent's lifetime (<i>note that there
 * is no 1:1 correspondance between the farm simulator's pulse given to plugins
 * and metagents, and the Agent's pulse function</i>).  It is the developer's 
 * responsiblity to ensure the agent periodically calls checkRun to ensure
 * that proper time allowances are respected.
 */
public class Agent extends Thread {

    // Status information
    private boolean stop = false;
    private volatile boolean running = false;

    // For time tracking
    private long lasttime = 0;
    private volatile long runtime = 0;
    private volatile boolean realtime = true;
    private boolean allowendrun = true;
    //private static JProf sysTimer = new JProf();

    // Locks
    Boolean runtimelock = new Boolean(true);
    Object metagentnotify = null;

    // For messages
    private Vector incoming = new Vector();
    private Vector outgoing = new Vector();

    // Utilities
    protected FarmMetAgent metagent = null;
    protected Log log = null;
    protected Random random = new Random();

    // Local information
    private String name = "Unknown";
    private int total = -1;

    /**
     * Constructor
     * @param metagent The meta-agent to be in control of this agent
     * @param name The name of the agent
     */
    public Agent(FarmMetAgent metagent, String name) {
        super();

        log = Log.getDefault();

        this.metagent = metagent;
        this.name = name;
        setName(name);
    }

    /**
     * Accessors
     */
    public String getAgentName() { return name; }

    /**
     * Adds additional runtime to the agent's allocated time.
     * @param time The amount of time to allow processing in milliseconds
     * @param real If true, real wall-clock time is used to track
     * execution.  If false, the time is interpreted as the number of
     * times checkRun may be called before halting.
     */
    public void setRun(long time, boolean real, Object notify) {

        // This effectively resets the local time counter each time 
        // time is added.  This is needed for some systems which don't
        // support our time monitoring scheme, but on systems that
        // work correctly this will make them less precise by hiding
        // extra time that might have been used in a previous pulse.
        runtime = getTotalRunTime();

        // Add the time and notify the thread
        if (time > 0) {
            realtime = real;
            runtime += time;
        }

        synchronized (runtimelock) {
            metagentnotify = notify;
            running = true;
            runtimelock.notify();
        }
    }

    /**
     * Stops the agent by artificially reducing its allocated runtime
     */
    public void endRun() {
        
	if (allowendrun) {
            if (realtime) {
                runtime = getTotalRunTime();
            } else {
                lasttime = runtime;
            }
	}
        checkRun();
    }

    /**
     * Determines if endRun calls are allowed or not.  If this
     * is set false, endRun calls will be ignored.
     */
    public void setAllowEndRun(boolean b) { allowendrun = b; }

    /**
     * Determines if the agent should continue running or not
     */
    private boolean shouldRun() {
    
        if (realtime) {
            lasttime = getTotalRunTime();
        } else {
            lasttime++;
        }

        //log.log("Current time: " + lasttime + " Stop time: " + runtime + " Running: " + running);
        return lasttime < runtime;
    }

    /**
     * Returns the total aggregate runtime of the agent in milliseconds.
     */
    private long getTotalRunTime() {

	// Under linux kernel 2.4.20-18.9 this function returns
	// timing information (units unknown), but does not appear
	// to differentiate btween threads.  All threads seem to
	// share the same common timer
        //log.log(getName() + " systimer: " + sysTimer.getThreadTime() + " " + Thread.currentThread().getName());
        //sysTimer.getThreadTime();

        return System.currentTimeMillis();
    }

    /**
     * Call this periodically to see if the agent should continue
     * its processing.  This will both check the runtime of the agent
     * and block until time is available.
     */
    public void checkRun() {

        // Check if we should be running
        while (!shouldRun()) {
            if (shouldStop()) {
                throw new IllegalStateException("Stop flag");
            }
            // Wait if we are done
            synchronized (runtimelock) {
                running = false;
                if (metagentnotify != null) {
                    synchronized (metagentnotify) {
                        metagentnotify.notify();
                    }
                    metagentnotify = null;
                }
                try {
                    runtimelock.wait();
                } catch (InterruptedException e) {
                    throw new IllegalStateException("Interrupted");
                }
            }
            if (shouldStop()) {
                throw new IllegalStateException("Stop flag");
            }
        }
    }

    /**
     * Determines if the agent should continue running or not
     */
    public boolean isRunning() {

        return running;
    }

    /**
     * If called, the "stop" flag is set for the agent, so that it
     * will terminate operations at the earliest convenient point.
     */
    public void setStop() {

        stop = true;
    }

    /**
     * Determines if the agent should stop or not.
     */
    protected boolean shouldStop() {

        return stop;
    }

    /**
     * Gets a global state property, using the owning metagent.
     */
    public Object getProperty(String name) {

        return metagent.getProperty(name);
    }

    /**
     * Sets a global state property, using the owning metagent.
     */
    public void setProperty(String name, Object value) {

        metagent.setProperty(name, value);
    }

    /**
     * Sends a message to another agent.  This function calls
     * prepareMessageDelivery to calculate the message's delivery
     * characteristics.
     * @param msg The message to send
     * @see FarmMetAgent#prepareMessageDelivery
     */
    public void sendMessage(Message msg) {

        synchronized (outgoing) {
            int time = getTime();
            
            // The agent believes it is being sent now
            msg.setSendTime(time);

            // Have the metagent determine delivery characteristics
            metagent.prepareMessageDelivery(this, msg);

            //log.log("Sending message during times: " + msg.getActualSendTime() + "-" + msg.getActualReceiveTime(), Log.LOG_INFO);
            log.log("Sending message " + msg, Log.LOG_INFO);
            outgoing.add(msg);
        }
    }

    /**
     * Returns, but does not remove the first message available
     * which was sent from the agent
     * @return A message, or null if none is avaiable
     */
    public Message peekFirstOutgoingMessage() {
        Message msg = null;

        synchronized (outgoing) {
            if (outgoing.size() > 0) {
                msg = (Message)outgoing.firstElement();
            }
        }

        return msg;
    }

    /**
     * Returns, but does not remove the last message available
     * which is to be sent from the agent
     * @return A message, or null if none is avaiable
     */
    public Message peekLastOutgoingMessage() {
        Message msg = null;

        synchronized (outgoing) {
            if (outgoing.size() > 0) {
                msg = (Message)outgoing.lastElement();
            }
        }

        return msg;
    }

    /**
     * Removes and returns the first message available
     * which was sent from the agent
     * @return A message, or null if none is avaiable
     */
    public Message getFirstOutgoingMessage() {
        Message msg = null;

        synchronized (outgoing) {
            if (outgoing.size() > 0) {
                msg = (Message)outgoing.firstElement();
                outgoing.removeElement(msg);
            }
        }

        return msg;
    }

    /**
     * Counts the number of messages which are waiting 
     * in the outgoing queue
     */
    public int countOutgoingMessages() {
        int size = 0;

        synchronized (outgoing) {
            size = outgoing.size();
        }

        return size;
    }

    /**
     * Recieves a messages from another agent
     * @param msg The message to receive
     */
    public void receiveMessage(Message msg) {

        synchronized (incoming) {
            msg.setReceiveTime(getTime());
            incoming.add(msg);
            log.log("Received message: " + msg, Log.LOG_INFO);
        }
    }

    /**
     * Removes and returns the first message available to
     * the agent.
     * @return A message, or null if none is avaiable
     */
    public Message getFirstIncomingMessage() {
        Message msg = null;

        synchronized (incoming) {
            if (incoming.size() > 0) {
                msg = (Message)incoming.firstElement();
                incoming.removeElement(msg);
            }
        }

        return msg;
    }

    /**
     * Returns, but does not remove, the first message available to
     * the agent.
     * @return A message, or null if none is avaiable
     */
    public Message peekFirstIncomingMessage() {
        Message msg = null;

        synchronized (incoming) {
            if (incoming.size() > 0) {
                msg = (Message)incoming.firstElement();
            }
        }

        return msg;
    }

    /**
     * Counts the number of messages which are waiting 
     * in the incoming queue
     */
    public int countIncomingMessages() {
        int size = 0;

        synchronized (incoming) {
            size = incoming.size();
        }

        return size;
    }

    /**
     * Returns true if there are incoming messages
     * awaiting processing
     * @return true when incoming messages exist
     */
    public boolean hasIncomingMessage() {

        return (countIncomingMessages() > 0);
    }

    /**
     * Convenience function, this just calls getProperty
     * to obtain the value of Simulator.P_TIME.
     */
    public int getTime() {
        Integer i = (Integer)getProperty(Simulator.P_TIME);
        return (i != null) ? i.intValue() : -1;
    }

    /**
     * The main control loop for the agent.  This essentially
     * just calls the pulse function repeatedly.
     */
    public void run() {

        begin();

        checkRun();

        try {
            while (!shouldStop()) {
                pulse();
            }
        } catch (RuntimeException e) {
            log.log("Ending, was interrupted by: " + e, Log.LOG_INFO);
            e.printStackTrace();
            setStop();
        }

        end();
    }

    /**
     * This is run when the agent first starts up
     */
    public void begin() {

        // Initialize the time variable
        lasttime = runtime = getTotalRunTime();

	// Get end run flag
	setAllowEndRun(metagent.getAllowEndRun());

        total = 0;

    }

    /**
     * This function gets called repeatedly during the execution
     * of the agent.  You should override this function in
     * derived classes to allow the agent to exhibit its specific
     * functionality.
     */
    public void pulse() {

        // Should we run?
        checkRun();

        // Increment the counter
        total++;
    }

    /**
     * This is run when the agent is stopped
     */
    public void end() {
        
        // We are no longer running
        running = false;

        if (total >= 0) {
            log.log("I was pulsed " + total + " times, total runtime is " + getTotalRunTime());
        }
    }
}
