/**
 */

package farm.simulator;

import java.rmi.*;
import java.rmi.server.*;
import java.util.*;
import java.io.*;

import farm.metagent.MetAgent;
import utilities.Log;

/**
 * The core simulation component used in farm.  This is responsible for
 * acting as a hub of plugin and metagent registration, maintaining the 
 * global property table, and enacting the control loop.  This component
 * must be started before all other plugins.
 * <P>
 * FarmSimulator supports the saving and replaying of state data over time
 * by using the -s <savefile> and -r <savefile> parameters respectively.
 * The minimum number of plugins may also be set on the command line with
 * the -m <minnumber> parameter.
 * <p>
 * Parameters:<dl>
 * <dt><tt>-s [filename]</tt>
 * <dd>If set, the state of the simulator is continuously saved in the
 * specified file
 *
 * <dt><tt>-r [filename]</tt>
 * <dd>If set, the simulator reads and applies the property states recorded
 * in the specified file
 *
 * <dt><tt>-m [integer]</tt>
 * <dd>If set, the simulator will not begin pulsing components until at least
 * the specified number of plugins have sucessfully connected.
 *
 * <dt><tt>-delay [integer]</tt>
 * <dd>The amount of ms to delay between subsequent pulse cycles (default 0)
 * 
 * <dt><tt>-readquit [boolean]</tt>
 * <dd>If true, the simulation will terminate after the end of a read save
 * file has been reached (default false)
 * 
 * <dt><tt>-propertyowners [string]</tt>
 * <dd>The name of the file (if any) in which to store the default property
 * owner information.  Files ending in .dat are expected to be a series
 * of serialized DataToken objects.  The final allocation will also be stored
 * in this file when the simulation terminates.  Files ending in .txt will
 * expect data in the form:
 *   OWNER\tProperty Name\tOwner PlugIn\tPushTarget1,PushTarget2,...
 * The final allocation will not be stored in .txt files.
 *
 * <dt><tt>-basecpu [double]</tt>
 * <dd>Baseline CPU performance metric, to be used with time scaling.  If
 * 0, the simulator will dynamically calculate its own metric (default 0).
 *
 * <dt><tt>-scaletime [boolean]</tt>
 * <dd>If true, the simulator will provide the agents with its baseline CPU
 * metric, which is then used by those agents along with their local metrics
 * to scale the pulse time values they are given (default false).
 * </dl>
 */
public class FarmSimulator extends UnicastRemoteObject implements Simulator {

    // Constants
    public static final int []SIMULATOR_PORTS = { 1099, 2105, 6667 };

    public static final String CLA_SAVESTATE = "s";
    public static final String CLA_READSTATE = "r";
    public static final String CLA_MINPLUGINS = "m";
    public static final String CLA_PULSEDELAY = "delay";
    public static final String CLA_READQUIT = "readquit";
    public static final String CLA_PROPOWNERS = "propertyowners";
    public static final String CLA_BASECPU = "basecpu";
    public static final String CLA_SCALETIME = "scaletime";

    // Remote objects
    ArrayList metagents = new ArrayList();
    Map plugins = new Hashtable();

    // Local information
    private Thread shutdownHook;
    private HashMap properties = new HashMap();
    private Hashtable propertytokens = new Hashtable();
    private Hashtable statictokens = new Hashtable();
    private Hashtable tokennotifys = new Hashtable();
    protected Log log = null;
    protected boolean done = false;
    protected boolean dead = false;
    protected int lasttime = 0;
    protected ObjectOutputStream savestream = null;
    protected ObjectInputStream readstream = null;
    protected int minplugins = 0;
    protected int pulsedelay = 0;
    protected boolean readquit = false;
    protected String propertyowners = null;
    protected double basecpu = 0;
    protected boolean scaletime = false;

    /**
     * Constructor.
     */
    public FarmSimulator(String []args) throws RemoteException {
        super();

        log = Log.getDefault();

        // Add the shutdown hook
        try {
            Runtime.getRuntime().addShutdownHook(shutdownHook = new Thread("ShutdownHook") {
                    public void run() {
                        log.log("Shutdown hook starting", Log.LOG_INFO);
                        setStop();
                        while (!dead) {
                            mysleep(100);
                        }
                        log.log("Shutdown hook completed", Log.LOG_INFO);
                    }
                });
        } catch (NoSuchMethodError ex) {
            log.log("Failed to addShutdownHook, you may be running an older version of java", Log.LOG_WARNING);
        }

        // Command line args
        startSave(getCommandLineArgument(args, CLA_SAVESTATE, null));
        startRead(getCommandLineArgument(args, CLA_READSTATE, null));
        setMinimumPlugIns(Integer.parseInt(getCommandLineArgument(args, CLA_MINPLUGINS, String.valueOf(getMinimumPlugIns()))));
        setPulseDelay(Integer.parseInt(getCommandLineArgument(args, CLA_PULSEDELAY, String.valueOf(getPulseDelay()))));
        setReadQuit(Boolean.valueOf(getCommandLineArgument(args, CLA_READQUIT, String.valueOf(getReadQuit()))).booleanValue());
        setScaleTime(Boolean.valueOf(getCommandLineArgument(args, CLA_SCALETIME, String.valueOf(getScaleTime()))).booleanValue());
        setBaselineCPUMetric(Double.parseDouble(getCommandLineArgument(args, CLA_BASECPU, String.valueOf(getBaselineCPUMetric()))));
        propertyowners = getCommandLineArgument(args, CLA_PROPOWNERS, propertyowners);

        // Determine baseline CPU performance if unknown
        if (scaletime && getBaselineCPUMetric() <= 0) {
            Linpack bench = new Linpack();
            setBaselineCPUMetric(bench.doBenchmark());
        }

        // Read the property owners
        if (propertyowners != null) {
            readPropertyOwners(propertyowners);
        }
    }

    /**
     * Command line argument parser.  This takes in the complete
     * list of CL arguments given to the process, the name of the
     * parameter to search for, and a default value to return
     * if that parameter is not found.  The parameter name
     * is prepended by a "-" during the search, so a given name
     * of "foo" will match "-foo" and "-foo bar".
     * @param args The argument list passed to the main function
     * @param name The name of the parameter to look for
     * @param def The default value of the parameter
     * @return The parameter as specified on the command line,
     *  or the default value if none found.  If a parameter
     *  has no value (e.g. the parameter name "-foo" is present
     *  but there is no value "bar") then the parameter name
     *  is returned.
     */
    public static String getCommandLineArgument(String []args, String name, String def) {

        for (int i = 0; i < args.length; i += 1) {
            if (args[i].equalsIgnoreCase("-" + name)) {
                if (args.length > i+1 && !args[i+1].startsWith("-")) {
                    return args[i+1];
                } else {
                    return name;
                }
            }
        }

        return def;
    }

    /**
     * Accessors
     */
    public void setPulseDelay(int i) { pulsedelay = i; }
    public int getPulseDelay() { return pulsedelay; }
    public double getBaselineCPUMetric() { return basecpu; }
    public void setBaselineCPUMetric(double b) {
        basecpu = b;
        log.log("Baseline CPU performance is " + basecpu, Log.LOG_INFO);
    }
    public void setReadQuit(boolean b) { readquit = b; }
    public boolean getReadQuit() { return readquit; }
    public void setScaleTime(boolean b) { scaletime = b; }
    public boolean getScaleTime() { return scaletime; }

    /**
     * If called, the "stop" flag is set for the plugin, so that it
     * will terminate operations at the earliest convenient point.
     */
    public void setStop() {

        if (isStopping()) {
            log.log("Smacking plugins around...", Log.LOG_INFO);
            //stopAllPlugIns();
        } else {

            log.log("Starting to stop, please wait...", Log.LOG_INFO);
        }

        done = true;
    }

    /**
     * Determines if the stop flag has been set or not
     */
    public boolean isStopping() {

        return done;
    }

    /**
     * If called, the "dead" flag is set for the plugin, declaring
     * the agent dead
     */
    public void setDead() {

        log.log("Simulator is dead", Log.LOG_INFO);
        dead = true;
    }

    /**
     * Sets a place to store property information during the run
     * @param name The name of the file to save the data to
     */
    public void startSave(String name) {
        if (name == null) return;

        try {
            FileOutputStream fstream = new FileOutputStream(name);
            savestream = new ObjectOutputStream(fstream);
            log.log("Saving property data to " + name, Log.LOG_INFO);
        } catch (FileNotFoundException e) { handleException(e); }
        catch (IOException e) { handleException(e); }
    }

    /**
     * Stops information saving
     */
    public void stopSave() {

        if (savestream != null) {
            synchronized (savestream) {
                try {
                    savestream.close();
                    savestream = null;
                    log.log("Stopping save of property data", Log.LOG_INFO);
                } catch (IOException e) { handleException(e); }
            }
        }
    }

    /**
     * Saves a particular object
     * @param object The object to save
     */
    protected void saveObject(Object object) {

        if (savestream != null) {
            synchronized (savestream) {
                try {
                    savestream.writeObject(object);
                } catch (IOException e) { handleException(e); }
            }
        }
    }

    /**
     * Sets a place to read property information from during the run
     * @param name The name of the file to read the data from
     */
    public void startRead(String name) {
        if (name == null) return;

        try {
            FileInputStream fstream = new FileInputStream(name);
            readstream = new ObjectInputStream(fstream);
            log.log("Reading property data from " + name, Log.LOG_INFO);
        } catch (FileNotFoundException e) { handleException(e); }
        catch (IOException e) { handleException(e); }
    }

    /**
     * Stops information reading
     */
    public void stopRead() {

        if (readstream != null) {
            synchronized (readstream) {
                try {
                    readstream.close();
                    readstream = null;
                    log.log("Stop read property data", Log.LOG_INFO);
                } catch (IOException e) { handleException(e); }
            }
            if (getReadQuit()) {
                setStop();
            }
        }
    }

    /**
     * Reads a particular object
     * @param object The object that was read, or null if none
     */
    protected Object readObject() {
        Object object = null;
        
        if (readstream != null) {
            synchronized (readstream) {
                try {
                    object = readstream.readObject();
                } catch (EOFException e) { stopRead(); }
                catch (IOException e) { handleException(e); }
                catch (ClassNotFoundException e) { handleException(e); }
            }
        }

        return object;
    }

    /**
     * Registers a metagent
     * @param agent The metagent in question
     */
    public void registerMetAgent(MetAgent agent) {

        synchronized (metagents) {
            try {
                log.log("Adding metagent " + agent.getMetAgentName(), Log.LOG_INFO);
            } catch (RemoteException e) { handleException(e); }

            metagents.add(agent);

            if (scaletime) {
                try {
                    agent.setBaselineCPUMetric(getBaselineCPUMetric());
                } catch (RemoteException e) { handleException(e); }
            }

            for (Iterator i = metagents.iterator(); i.hasNext();) {
                MetAgent metagent = (MetAgent)i.next();
                if (metagent != agent) {
                    try {
                        metagent.registerMetAgent(agent);
                    } catch (RemoteException e) { handleException(e); }
                    try {
                        agent.registerMetAgent(metagent);
                    } catch (RemoteException e) { handleException(e); }
                }
            }
        }
    }

    /**
     * Unregisters a metagent
     * @param agent The metagent in question
     */
    public void unregisterMetAgent(MetAgent agent) {

        synchronized (metagents) {
            try {
                log.log("Removing metagent " + agent.getMetAgentName(), Log.LOG_INFO);
            } catch (RemoteException e) { handleException(e); }
            metagents.remove(agent);

            for (Iterator i = metagents.iterator(); i.hasNext();) {
                MetAgent metagent = (MetAgent)i.next();
                if (metagent != agent) {
                    try {
                        metagent.unregisterMetAgent(agent);
                    } catch (RemoteException e) { handleException(e); }
                }
            }
        }
    }

    /**
     * Gets an iterator over the metagents
     */
    public Iterator getMetAgents() {
        Vector v = new Vector();

        synchronized (metagents) {
            int size = metagents.size();
            for (int i = 0; i < size; i++) {
                MetAgent metagent = (MetAgent)metagents.get(i);
                v.addElement(metagent);
            }            
        }

        return v.iterator();
    }

    /**
     * Registers a plugin
     * @param plugin The plugin in question
     */
    public void registerPlugIn(PlugIn plugin) {

        try {
            String pluginname = plugin.getPlugInName();
            log.log("Adding plugin " + pluginname + ": " + plugin, Log.LOG_INFO);

            // Add them
            synchronized (plugins) {
                if (plugins.containsKey(pluginname)) {
                    log.log("Warning: plugin " + pluginname + " is already registered");
                    return;
                }
                plugins.put(pluginname, plugin);
            }

        } catch (RemoteException e) { handleException(e); }
    }

    /**
     * Unregisters a plugin
     * @param plugin The plugin in question
     */
    public void unregisterPlugIn(PlugIn plugin) {

        try {
            String pluginname = plugin.getPlugInName();
            log.log("Removing plugin " + pluginname, Log.LOG_INFO);

            // Unregister the properties they own
            Vector props = new Vector();
            synchronized (propertytokens) {
                for (Iterator i = propertytokens.keySet().iterator(); i.hasNext(); ) {
                    String name = (String)i.next();
                    DataToken token = (DataToken)propertytokens.get(name);
                    if (token.getSource() != null &&
                        token.getSource().equals(pluginname)) {
                        props.add(name);
                    }
                }
            }
            for (Iterator i = props.iterator(); i.hasNext(); ) {
                String name = (String)i.next();
                unsetPropertyOwner(name, pluginname);
            }

            synchronized (plugins) {
                plugins.remove(pluginname);
            }

        } catch (RemoteException e) { handleException(e); }
    }

    /**
     * Returns the number of plugins that are registered
     */
    public int countPlugIns() {
        int count = 0;

        synchronized (plugins) {
            count = plugins.size();
        }

        return count;
    }

    /**
     * Registers a plugin
     * @param plugin The plugin in question
     */
    public PlugIn findPlugIn(String name) {

        synchronized (plugins) {
            return (PlugIn)plugins.get(name);
        }
    }

    /**
     * Gets an iterator over the plugins
     */
    public Iterator getPlugIns() {
        Collection v = new ArrayList(25);

        synchronized (plugins) {
            for (Iterator i = plugins.entrySet().iterator(); i.hasNext(); ) {
                Map.Entry entry = (Map.Entry)i.next();
                v.add((PlugIn)entry.getValue());
            }            
        }

        return v.iterator();
    }

    /**
     * Gets an iterator over the plugins
     */
    public Collection getPlugInNames() {
        Collection v = new ArrayList();

        synchronized (plugins) {
            for (Iterator i = plugins.keySet().iterator(); i.hasNext(); ) {
                v.add((String)i.next());
            }            
        }

        return v;
    }

    /**
     * From a plugin reference, determine it's name
     * @param plugin The plugin in question
     */
    public String getPlugInName(PlugIn plugin) {
        String name = "???";

        try {
            name = plugin.getPlugInName();
        } catch (RemoteException ex) { 
            log.log("Error directly getting plugin name, searching table: " + ex);
            synchronized (plugins) {
                for (Iterator i = plugins.keySet().iterator(); i.hasNext(); ) {
                    String key = (String)i.next();
                    if (plugins.get(key) == plugin) {
                        name = key;
                        break;
                    }
                }
            }
        }

        return name;
    }

    /**
     * Stops and removes all plugins
     */
    public void stopAllPlugIns() {

	// Tell the agents to stop
	//synchronized (plugins) {
	for (Iterator i = getPlugIns(); i.hasNext();) {
	    try {
                PlugIn plugin = (PlugIn)i.next();
                try {
                    plugin.setStop();
                } catch (RemoteException ex) { handleException(ex); }
	    } catch (NoSuchElementException ex) {
		log.log("Tried to stop a plugin that's already gone.", Log.LOG_INFO);
		log.log(ex.toString(), Log.LOG_INFO);
	    }
	}
	//}

	// Wait for them to complete
	for (int x = 1; countPlugIns() > 0 && x < 100 * 10; x++) {
	    mysleep(50);
	    if (x % 100 == 0) {
		String slackers = "";
		for (Iterator i = getPlugIns(); i.hasNext();) {
		    PlugIn plugin = (PlugIn)i.next();
                    slackers += " " + getPlugInName(plugin);
		    try {
			plugin.setStop();
		    } catch (RemoteException ex) {
			slackers += "(error)";
			handleException(ex);
		    }
		}
		log.log("Waiting for:" + slackers, Log.LOG_INFO);
		System.out.println("Waiting for: " + slackers);
	    }
	}
    }
    
    /**
     * Specifies a minimum number of plugins which must be
     * registered before the system will begin pulsing.  Note
     * this is only currently checked in the begin phase of
     * simulator start-up.
     * @param m The minimum number of plugins the simulator
     *   wants before starting.
     */
    public void setMinimumPlugIns(int m) {

        minplugins = m;
    }

    /**
     * Returns the minimum start number of plugins.
     * @see #setMinimumPlugIns
     */
    public int getMinimumPlugIns() {
        
        return minplugins;
    }
    
    /**
     * Handles a local property request.
     * @param name The name of the property to access
     * @return The object, or null if not found
     */
    public Object getProperty(String name) {
        Object value = null;

        // If it has a token, check it
        DataToken token = null;
        if ((token = (DataToken)propertytokens.get(name)) != null) {
            // If I'm the owner, or the owner hasn't been told, I should have it
            if (token.isOwner() ||
                tokennotifys.get(name) != Boolean.TRUE) {
                value = getLocalProperty(name);
            // If someone else is the owner, redirect them
            } else if (token.getSource() != null) {
                PlugIn plugin = token.getSourcePlugIn();
                if (plugin == null) {
                    plugin = findPlugIn(token.getSource());
                    token.setSourcePlugIn(plugin);
                }
                try {
                    value = plugin.getPropertyRemote(name);
                } catch (RemoteException e) { handleException(e); }
            // Bad token, bad!
            } else {
                log.log("Illegal data token for property '" + name + "': " + token);
            }
        } else {
            value = getLocalProperty(name);
        }

        return value;
    }

    /**
     * Gets the local property value
     * @param name The name of the property to access
     */
    public Object getLocalProperty(String name) {

        synchronized (properties) {
            return properties.get(name);
        }
    }

    /**
     * Handles a remote property request.
     * @param name The name of the property to access
     * @return The object, or null if not found
     */
    public Object getPropertyRemote(String name) {
        Object value = null;

        // If it has a token, check it
        DataToken token = null;
        if ((token = (DataToken)propertytokens.get(name)) != null) {
            // If I'm the owner, I should have it
            if (token.isOwner() ||
                tokennotifys.get(name) != Boolean.TRUE) {
                value = properties.get(name);
            // If someone else is the owner, redirect them
            } else if (token.getSource() != null &&
                       findPlugIn(token.getSource()) != null) {
                value = getGenericPropertyToken(token.getSource());
            // Bad token, bad!
            } else {
                log.log("Illegal data token for property '" + name + "': " + token);
            }
        }
        if (value == null) {
            value = getLocalProperty(name);
        }

        return value;
    }

    /**
     * Handles a local property set command. 
     * @param name The name of the property to access
     * @param value The value to set it to
     */
    public boolean setProperty(String name, Object value) {

        // If it has a token, check it
        if (propertytokens.containsKey(name) &&
            tokennotifys.get(name) == Boolean.TRUE) {
            DataToken token = (DataToken)propertytokens.get(name);
            // If someone else is the owner, redirect them
            if (token.getSource() != null) {
                PlugIn plugin = token.getSourcePlugIn();
                if (plugin == null) {
                    plugin = findPlugIn(token.getSource());
                    token.setSourcePlugIn(plugin);
                }
                try {
                    plugin.setPropertyRemote(name, value);
                } catch (RemoteException e) {
                    e.printStackTrace();
                }
            }
        } else {
            setLocalProperty(name, value);
        }

        return true;
    }

    /**
     * Handles a property set command. 
     * @param name The name of the property to access
     * @param value The value to set it to
     * @return null if the property was set, a DataToken redirect otherwise
     */
    public DataToken setPropertyRemote(String name, Object value) {

        //System.err.println("Setting " + name + " to " + value + " (mem " + Runtime.getRuntime().freeMemory() + ")");
        
        // Save it regardless
        setLocalProperty(name, value);

        // If it has a token, check it
        if (propertytokens.containsKey(name) &&
            tokennotifys.get(name) == Boolean.TRUE) {
            DataToken token = (DataToken)propertytokens.get(name);
            // If someone else is the owner, redirect them
            if (token.getSource() != null &&
                findPlugIn(token.getSource()) != null) {
                return getGenericPropertyToken(token.getSource());
            }
        }

        //System.err.println("Setting " + name + " to " + value + " (mem " + Runtime.getRuntime().freeMemory() + ")");

        return null;
    }

    /**
     * Sets a local property.
     * @param name The name of the property to access
     * @param value The data to store there
     */
    public void setLocalProperty(String name, Object value) {
        DataToken token = null;

        synchronized (properties) {
            if (value != null) {
                properties.put(name, value);
            } else {
                properties.remove(name);
            }
        }

        if (savestream != null) {
            saveObject(name);
            saveObject(value);
        }
    }

    /**
     * Returns a generic property token for a particular
     * source.  These should be consider immutable.
     * @return A DataToken
     */
    public DataToken getGenericPropertyToken(String source) {
        DataToken token = (DataToken)statictokens.get(source);

        if (token == null) {
            token = new DataToken(source);
            statictokens.put(source, token);
        }

        return token;
    }

    /**
     * Reads in data from the property owners file
     * @param filename The name of the owners file
     */
    public void readPropertyOwners(String filename) {

        log.log("Reading property owner information from " + filename);
        long stime = System.currentTimeMillis();
        try {
            if (filename.endsWith(".dat")) {
                FileInputStream file = new FileInputStream(filename);
                ObjectInputStream stream = new ObjectInputStream(file);
                while (true) {
                    Object o1 = stream.readObject();
                    Object o2 = stream.readObject();
                    if (o1 instanceof String &&
                        o2 instanceof DataToken) {
                        String name = (String)o1;
                        DataToken token = (DataToken)o2;
                        setPropertyOwner(name, token.getSource(), token.getCacheTargetsCollection());
                    }
                }
            } else if (filename.endsWith(".txt")) {
                FileReader file = new FileReader(filename);
                BufferedReader reader = new BufferedReader(file);
                String str;
                while ((str = reader.readLine()) != null) {
                    StringTokenizer tok = new StringTokenizer(str, "\t");
                    tok.nextToken();
                    String name = tok.nextToken();
                    String source = tok.nextToken();
                    String pushstr = "";
                    if (tok.hasMoreTokens())
                        pushstr = tok.nextToken();
                    Collection pushees = new ArrayList();
                    StringTokenizer tok2 = new StringTokenizer(pushstr, ",");
                    while (tok2.hasMoreTokens()) {
                        pushees.add(tok2.nextToken());
                    }
                    setPropertyOwner(name, source, pushees);
                }
            }

        } catch (EOFException e) {
            // Done
        } catch (Exception e) {
            handleException(e);
        }
        System.err.println("Read owners in " + (System.currentTimeMillis() - stime));
    }

    /**
     * Reads in data from the property owners file
     * @param filename The name of the owners file
     */
    public void writePropertyOwners(String filename) {
        if (!filename.endsWith(".dat")) return;

        log.log("Saving property owner information to " + filename);
        try {
            FileOutputStream file = new FileOutputStream(filename);
            ObjectOutputStream stream = new ObjectOutputStream(file);
            try {
                synchronized (propertytokens) {
                    for (Iterator i = propertytokens.keySet().iterator(); i.hasNext(); ) {
                        String name = (String)i.next();
                        DataToken token = (DataToken)propertytokens.get(name);

                        if (token.getSource() != null) {
                            try {
                                PlugIn plugin = findPlugIn(token.getSource());
                                DataToken remotetoken = plugin.getPropertyToken(name);
                                if (remotetoken != null) {
                                    token.setCacheTargets(remotetoken.getCacheTargetsCollection());
                                }
                            } catch (RemoteException e) { handleException(e); }
                        }

                        stream.writeObject(name);
                        System.err.println("Writing " + name + "\t" + token);
                        stream.writeObject(token);
                    }
                }
            }
            catch (IOException e) { handleException(e); }

        } catch (Exception e) {
            handleException(e);
        }
    }

    /**
     * Sets the owner of a property data item.  This is done
     * by first disowning the current owner (if any), and then
     * notifying the new owner of its status.
     * @param name The name of the property
     * @param source The plugin's new owner
     * @param pushees The targets the owner should push the value to
     */
    public synchronized void setPropertyOwner(String name, String source, Collection pushees) {

        // Remove previous owner
        DataToken token = (DataToken)propertytokens.get(name);
        if (token != null) {
            if (pushees == null) {
                pushees = token.getCacheTargetsCollection();
            }
            if (token.getSource() != null &&
                !token.getSource().equals(source)) {
                unsetPropertyOwner(name, token.getSource());
            }
        }

        // Set new owner
        synchronized (tokennotifys) {
            try {
                // Notify the plugin if we can
                PlugIn plugin = (PlugIn)findPlugIn(source);
                if (plugin != null) {
                    Object value = properties.get(name);
                    plugin.ownProperty(name, value, pushees);
                    tokennotifys.put(name, Boolean.TRUE);
                    properties.put(name, null);

                    log.log("Plugin " + source + " now owns property " + name);

                } else {
                    log.log("Could not find plugin " + source + " to own property " + name);
                }
                
            } catch (RemoteException e) { handleException(e); }
        }

        // Save the token regardless
        synchronized (propertytokens) {
            if (pushees == null || pushees.isEmpty()) {
                propertytokens.put(name, getGenericPropertyToken(source));
            } else {
                token = new DataToken(source);
                token.addCacheTargets(pushees);
                propertytokens.put(name, token);
            }
        }
    }

    /**
     * This revoked the ownership rights a plugin has over
     * a particular property.  This is done by first determining
     * that the particular plugin is in fact the correct owner.
     * The plugin is then disowned from the property and the
     * current value saved locally.  Finally, the local data token
     * is then removed.
     * @param name The name of the property
     * @param source The plugin to revoke ownership from
     */
    public synchronized void unsetPropertyOwner(String name, String source) {

        DataToken token = (DataToken)propertytokens.get(name);
        synchronized (tokennotifys) {
            if (token != null &&
                token.getSource().equals(source)) {
                try {
                    PlugIn plugin = (PlugIn)findPlugIn(source);
                    if (plugin != null) {
                        // Get the most recent instance of the variable
                        Object value = plugin.disownProperty(name);
                        tokennotifys.remove(name);
                        properties.put(name, value);

                    } else {
                        log.log("Could not find plugin " + source + " to disown property " + name);
                    }
                } catch (RemoteException e) { handleException(e); }
            } else {
                log.log("Plugin " + source + " was not the owner of property " + name);
            }

            synchronized (propertytokens) {
                propertytokens.remove(name);
                tokennotifys.remove(name);
            }
        }
    }

    /**
     * Finds property names which match the provided name.
     * @param name The name of the property to search for
     */
    public Collection findProperties(String name) {
        ArrayList matches = new ArrayList();
		
        for (Iterator i = properties.keySet().iterator(); i.hasNext(); ) {
            String s = (String)i.next();
            if (s.indexOf(name) >= 0) {
                matches.add(s);
            }
        }

        return matches;
    }

    /**
     * Finds local property names which match the provided name.
     * @param name The name of the property to search for
     */
    public Collection findLocalProperties(String name) {
        ArrayList matches = new ArrayList();
		
        for (Iterator i = properties.keySet().iterator(); i.hasNext(); ) {
            String s = (String)i.next();
            if (s.indexOf(name) >= 0 &&
                properties.get(s) != null) {
                matches.add(s);
            }
        }

        return matches;
    }

    /**
     * Runs a function locally at the simulator, useful
     * for functions that require excessive bandwidth
     * @param function The function to perform locally at
     * the simulator
     * @return Returns whatever object is returned by
     *  the LocalFunction that is passed in.
     */
    public Object runLocalFunction(LocalFunction function) throws RemoteException {

        return function.run(this);
    }

    /**
     * The main control loop
     */
    public void run() {

        // Control loop
        try {
            begin();
            while (!isStopping()) {
                long stime = System.currentTimeMillis();
		pulse();
                log.log("Pulse finished, " + (System.currentTimeMillis() - stime) + " ms elapsed", Log.LOG_INFO);
                mysleep(getPulseDelay());
            }
            end();

        } catch (Exception e) {
            handleException(e);
        } finally {
            if (!isStopping())
                setStop();
            setDead();
        }

        // Finish
        System.exit(0);
    }

    /**
     * Called periodically by the main control loop.  This works
     * by first sending a PULSE_BEGIN sequentially to all registered
     * metagents and plugins.  It then sends a PULSE_IN to all metagents
     * in  parallel, passing in the difference in the P_TIME property from
     * the last pulse to this one.  Finally, it sends a PULSE_END to all
     * metagents and plugins.
     * <p>
     * Note the simulator expects some other component to control the
     * granularity of the process by modifying the P_TIME property.
     */
    public void pulse() {
        ThreadGroup mathreads = new ThreadGroup("MetAgents");
	long stime = System.currentTimeMillis();

        // Read in saved data
        for (Object object = readObject(); object != null; object = readObject()) {
            if (object instanceof PulseEnd) {
                break;
            }
            Object value = readObject();
            if (value != null) {
                setProperty(object.toString(), value);
            }
        }

        // Start the pulse (plugins first)
        for (Iterator i = getPlugIns(); i.hasNext(); ) {
            PlugIn plugin = (PlugIn)i.next();
            if (!metagents.contains(plugin)) {
                pulsePlugIn(plugin, PULSE_BEGIN, 0);
            }
        }
        for (Iterator i = getMetAgents(); i.hasNext(); ) {
            PlugIn plugin = (PlugIn)i.next();
            pulsePlugIn(plugin, PULSE_BEGIN, 0);
        }
        /*for (Iterator i = getMetAgents(); i.hasNext(); ) {
            Thread t = new MetAgentThread(this, mathreads, (MetAgent)i.next(), PULSE_BEGIN, 0);
            t.start();
        }
        while (mathreads.activeCount() > 0) {
            mysleep(50);
        }*/
	log.log("Pulse begin finished, " + (System.currentTimeMillis() - stime) + " ms elapsed", Log.LOG_INFO);
	stime = System.currentTimeMillis();

        // What time is it?
        Integer time = (Integer)getProperty(Simulator.P_TIME);
        int thistime = (time != null) ? time.intValue() : 0;

        // Pulse the agents
        int timedelta = thistime - lasttime;
        long startTime = System.currentTimeMillis();
	for (Iterator i = getMetAgents(); i.hasNext(); ) {
            Thread t = new MetAgentThread(this, Thread.currentThread(), mathreads, (MetAgent)i.next(), PULSE_IN, timedelta);
            t.start();
        }
        while (MetAgentThread.count > 0) {
	    // If the agents complete after the above statement,
	    // but before the below statement, the simulator will
	    // sleep timedelta ms unecessarily.
	    mysleep(timedelta);
        }
	long runTime = System.currentTimeMillis() - startTime;
	setProperty("pulseLen", new Integer((int) runTime));

        // Save the time
        lasttime = thistime;

	log.log("Pulse in finished, " + (System.currentTimeMillis() - stime) + " ms elapsed", Log.LOG_INFO);
	stime = System.currentTimeMillis();

        // End the pulse (metagents first)
        /*for (Iterator i = getMetAgents(); i.hasNext(); ) {
            Thread t = new MetAgentThread(this, mathreads, (MetAgent)i.next(), PULSE_END, 0);
            t.start();
        }
        while (mathreads.activeCount() > 0) {
            mysleep(50);
        }*/
        for (Iterator i = getMetAgents(); i.hasNext(); ) {
            PlugIn plugin = (PlugIn)i.next();
            pulsePlugIn(plugin, PULSE_END, 0);
        }
        for (Iterator i = getPlugIns(); i.hasNext(); ) {
            PlugIn plugin = (PlugIn)i.next();
            if (!metagents.contains(plugin)) {
                pulsePlugIn(plugin, PULSE_END, 0);
            }
        }

        // Save a pulse end
        saveObject(new PulseEnd());

	log.log("Pulse end finished, " + (System.currentTimeMillis() - stime) + " ms elapsed", Log.LOG_INFO);
    }

    /**
     * Sends a pulse of a particular type to a plugin
     * @param plugin The plugin to pulse
     * @param pulse The pulse type id
     * @param duration The duration of the pulse (for PULSE_IN)
     */
    protected void pulsePlugIn(PlugIn plugin, int pulse, int duration) {

        try {
            long stime = System.currentTimeMillis();
            String name = getPlugInName(plugin);
            log.log("Pulsing " + name + ": " + pulse + " " + duration, Log.LOG_INFO);
            plugin.setPulse(pulse, duration);
            log.log("Done pulsing " + name + ", " + (System.currentTimeMillis() - stime) + " ms elapsed", Log.LOG_INFO);
        } catch (RemoteException e) {
            System.err.println("Exception: " + e.getMessage());
            e.printStackTrace();
            
            if (plugin instanceof MetAgent) {
                unregisterMetAgent((MetAgent)plugin);
            }
            unregisterPlugIn(plugin);
        }
    }

    /**
     * Called when the simulator starts.  This currently pauses until
     * the minimum number of plugins have registered before continuing.
     */
    public void begin() {        

        // Time begins
        lasttime = 0;

        // Wait
	int count = 1;
        long stime = System.currentTimeMillis();
        while (countPlugIns() < minplugins &&
               !isStopping()) {
            mysleep(1000);
            log.log("Waiting...have " + countPlugIns() + " of " + minplugins + " plugins", Log.LOG_INFO);
	    if (count++ % 10 == 0) {
		String plugs = "";
		for (Iterator i = getPlugIns(); i.hasNext();) {
		    PlugIn plugin = (PlugIn)i.next();
                    plugs += " " + getPlugInName(plugin);
		}
		log.log("Currently have:" + plugs, Log.LOG_INFO);
	    }
        }

        // Notify about properties
        if (!isStopping()) {
            for (Iterator names = getPlugInNames().iterator(); names.hasNext(); ) {
                String pluginname = (String) names.next();

                // Register the properties they own
                HashMap props = new HashMap();
                synchronized (propertytokens) {
                    for (Iterator i = propertytokens.keySet().iterator(); i.hasNext(); ) {
                        String name = (String)i.next();
                        DataToken token = (DataToken)propertytokens.get(name);
                        if (token.getSource() != null &&
                            token.getSource().equals(pluginname) &&
                            tokennotifys.get(name) != Boolean.TRUE) {
                            props.put(name, token.getCacheTargetsCollection());
                        }
                    }
                }
                for (Iterator i = props.keySet().iterator(); i.hasNext(); ) {
                    String name = (String)i.next();
                    Collection pushees = (Collection)props.get(name);
                    setPropertyOwner(name, pluginname, pushees);
                }
            }
        }

        log.log("Done...have " + countPlugIns() + " of " + minplugins + " plugins," + (System.currentTimeMillis() - stime) + " ms elapsed", Log.LOG_INFO);
    }

    /**
     * Called when the simulator is stopping.
     */
    public void end() {

        // Save properties
        if (propertyowners != null) {
            writePropertyOwners(propertyowners);
        }

	log.log("Stopping simulation...", Log.LOG_INFO);

        // End everything
        stopAllPlugIns();

        // Stop information processing
        stopRead();
        stopSave();
    }

    /**
     * Convenience function.
     */
    public void mysleep(int i) {
        if (i <= 0) return;

        try {
            Thread.currentThread().yield();
            Thread.currentThread().sleep(i);
        } catch (InterruptedException e) { }
    }

    /**
     * Convenience function.
     */
    protected void handleException(Exception e) {
        System.err.println("Exception in simulator: " + e.getMessage());
        e.printStackTrace();
    }

    /**
     */
    public static void main(String[] args) {

        // Set the security manager
        if (System.getSecurityManager() == null) {
            System.setSecurityManager(new RMISecurityManager());
        }

        // Create the simulator service
        FarmSimulator simulator = null;
        try {
            simulator = new FarmSimulator(args);
            for (int i = 0; i < SIMULATOR_PORTS.length; i++) {
                String host = getCommandLineArgument(args, CLA_SIMULATORHOST, "127.0.0.1");
                int port = SIMULATOR_PORTS[i];
                if (i == 0)
                    port = Integer.parseInt(getCommandLineArgument(args, CLA_SIMULATORPORT, String.valueOf(port)));
                try {
                    String name = "//" + host + ":" + port + "/simulator";

                    // Start it up
                    Naming.rebind(name, simulator);
                    simulator.run();
                    Naming.unbind(name);
                    break;

                } catch (ConnectException e) {
                    System.err.println("Error binding to rmiregistry on host " + host + " port " + port + ((i+1 < SIMULATOR_PORTS.length) ? ", trying port " + SIMULATOR_PORTS[i+1] : ", no more ports to try"));
                }
            }
        } catch (Error e) {
            System.err.println("Error: " + e.getMessage());
            e.printStackTrace();
        } catch (Exception e) {
            System.err.println("Exception: " + e.getMessage());
            e.printStackTrace();
        }

        if (simulator != null) {
            simulator.setDead();
        }

        System.exit(0);
    }
}

/**
 * Class for managing metagent activities
 */
class MetAgentThread extends Thread {
    MetAgent agent;
    FarmSimulator simulator;
    int type;
    int time;
    Thread master;
    static int count = 0;

    public MetAgentThread(FarmSimulator sim, Thread master, ThreadGroup group, MetAgent agent, int type, int time) {
        super(group, "MetAgent Thread");

	count++;
        this.simulator = sim;
        this.agent = agent;
        this.type = type;
        this.time = time;
	this.master = master;
    }

    public void run() {        
        simulator.pulsePlugIn(agent, type, time);
	count--;
	master.interrupt();
    }
}

/**
 * An object save file delimiter
 */
class PulseEnd implements Serializable {
}
